# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.utils.misc import assert_interfaces_valid
from qwt import QwtPlotItem

from plotpy.coords import canvas_to_axes
from plotpy.interfaces import IBasePlotItem, IShapeItemType

if TYPE_CHECKING:
    from qtpy.QtCore import QPointF  # helping out python_qt_documentation

    from plotpy.interfaces import IItemType
    from plotpy.plot import BasePlot
    from plotpy.styles.base import ItemParameters


class AbstractShape(QwtPlotItem):
    """Abstract shape class"""

    __implements__ = (IBasePlotItem,)

    _readonly = False
    _private = False
    _can_select = True
    _can_resize = True
    _can_rotate = False  # TODO: Implement shape rotation?
    _can_move = True

    def __init__(self) -> None:
        super().__init__()
        self.selected = False

    # ------IBasePlotItem API----------------------------------------------------
    def set_selectable(self, state: bool) -> None:
        """Set item selectable state

        Args:
            state: True if item is selectable, False otherwise
        """
        self._can_select = state

    def set_resizable(self, state: bool) -> None:
        """Set item resizable state
        (or any action triggered when moving an handle, e.g. rotation)

        Args:
            state: True if item is resizable, False otherwise
        """
        self._can_resize = state

    def set_movable(self, state: bool) -> None:
        """Set item movable state

        Args:
            state: True if item is movable, False otherwise
        """
        self._can_move = state

    def set_rotatable(self, state: bool) -> None:
        """Set item rotatable state

        Args:
            state: True if item is rotatable, False otherwise
        """
        self._can_rotate = state

    def can_select(self) -> bool:
        """
        Returns True if this item can be selected

        Returns:
            bool: True if item can be selected, False otherwise
        """
        return self._can_select

    def can_resize(self) -> bool:
        """
        Returns True if this item can be resized

        Returns:
            bool: True if item can be resized, False otherwise
        """
        return self._can_resize

    def can_rotate(self) -> bool:
        """
        Returns True if this item can be rotated

        Returns:
            bool: True if item can be rotated, False otherwise
        """
        return self._can_rotate

    def can_move(self) -> bool:
        """
        Returns True if this item can be moved

        Returns:
            bool: True if item can be moved, False otherwise
        """
        return self._can_move

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IShapeItemType,)

    def set_readonly(self, state: bool) -> None:
        """Set object readonly state

        Args:
            state: True if object is readonly, False otherwise
        """
        self._readonly = state

    def is_readonly(self) -> bool:
        """Return object readonly state

        Returns:
            bool: True if object is readonly, False otherwise
        """
        return self._readonly

    def set_private(self, state: bool) -> None:
        """Set object as private

        Args:
            state: True if object is private, False otherwise
        """
        self._private = state

    def is_private(self) -> bool:
        """Return True if object is private

        Returns:
            bool: True if object is private, False otherwise
        """
        return self._private

    def select(self) -> None:
        """
        Select the object and eventually change its appearance to highlight the
        fact that it's selected
        """
        self.selected = True
        self.invalidate_plot()

    def unselect(self) -> None:
        """
        Unselect the object and eventually restore its original appearance to
        highlight the fact that it's not selected anymore
        """
        self.selected = False
        self.invalidate_plot()

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        pass

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        pass

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """

    def set_item_parameters(self, itemparams):
        """
        Change the appearance of this item according
        to the parameter set provided

        params is a list of Datasets of the same types as those returned
        by get_item_parameters
        """
        pass

    def move_local_point_to(
        self, handle: int, pos: QPointF, ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        pt = canvas_to_axes(self, pos)
        self.move_point_to(handle, pt, ctrl)
        plot: BasePlot = self.plot()
        if plot is not None:
            plot.SIG_ITEM_RESIZED.emit(self, 0, 0)
            plot.SIG_ITEM_HANDLE_MOVED.emit(self)

    def move_local_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in canvas coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        old_pt = canvas_to_axes(self, old_pos)
        new_pt = canvas_to_axes(self, new_pos)
        self.move_shape(old_pt, new_pt)
        plot: BasePlot = self.plot()
        if plot is not None:
            plot.SIG_ITEM_MOVED.emit(self, *(old_pt + new_pt))

    def move_with_selection(self, delta_x: float, delta_y: float) -> None:
        """Translate the item together with other selected items

        Args:
            delta_x: Translation in plot coordinates along x-axis
            delta_y: Translation in plot coordinates along y-axis
        """
        self.move_shape([0, 0], [delta_x, delta_y])

    # ------Public API-----------------------------------------------------------
    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        pass

    def move_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        pass

    def invalidate_plot(self) -> None:
        """Invalidate the plot to force a redraw"""
        plot: BasePlot = self.plot()
        if plot is not None:
            plot.invalidate()

    def is_empty(self) -> bool:
        """Return True if the item is empty

        Returns:
            True if the item is empty, False otherwise
        """
        return False


assert_interfaces_valid(AbstractShape)



from __future__ import annotations

import numpy as np
import qtpy.QtCore as QC
from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.tests.unit.utils import (
    CLICK,
    create_window,
    drag_mouse,
    keyboard_event,
    mouse_event_at_relative_plot_pos,
)
from plotpy.tools import MultiLineTool, PolygonTool


def __generic_polyline_tool_test(
    toolklass: type[MultiLineTool] | type[PolygonTool], points: np.ndarray
) -> None:
    """Generic polyline tool test."""
    with qt_app_context(exec_loop=False):
        win, tool = create_window(toolklass)

        x0, y0 = None, None
        for x, y in points:
            if x0 is not None:
                x_path = np.linspace(x0, x, 10)
                y_path = np.linspace(y0, y, 10)
                drag_mouse(win, x_path, y_path)
            else:
                mouse_event_at_relative_plot_pos(win, (x, y), CLICK)
            x0, y0 = x, y

        assert tool.handler.shape is not None
        assert tool.handler.shape.get_points().shape == points.shape

        keyboard_event(win, QC.Qt.Key.Key_Backspace)

        points_count, _ = tool.handler.shape.get_points().shape

        assert points_count == (len(points) - 1)

        # add last point by dragging mouse
        drag_mouse(win, points[-2:, 0], points[-2:, 1])

        points_count, _ = tool.handler.shape.get_points().shape
        assert points_count == len(points)

        keyboard_event(win, QC.Qt.Key.Key_Space)

        exec_dialog(win)


def test_polygon_tool():
    """Test the polygon tool."""
    corners = np.array(((0.1, 0.1), (0.1, 0.8), (0.8, 0.8), (0.8, 0.1)))
    __generic_polyline_tool_test(PolygonTool, corners)


def test_multiline_tool():
    """Test the multi line tool."""
    n = 100
    t = np.linspace(0, np.pi * 10, n)
    x_arr = t * np.cos(t) / n + 0.5
    y_arr = t * np.sin(t) / n + 0.5
    __generic_polyline_tool_test(MultiLineTool, np.array([x_arr, y_arr]).T)


if __name__ == "__main__":
    test_polygon_tool()
    test_multiline_tool()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Testing image coordinates issues

Check that the first image pixel is centered on (0, 0) coordinates.

See https://github.com/PlotPyStack/guiqwt/issues/90
"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import data as ptd
from plotpy.tests import vistools as ptv
from plotpy.tools import DisplayCoordsTool


def test_pixel_coords():
    """Testing image pixel coordinates"""
    title = test_pixel_coords.__doc__
    data = ptd.gen_2d_gaussian(20, np.uint8, x0=-10, y0=-10, mu=7, sigma=10.0)
    with qt_app_context(exec_loop=True):
        image = make.image(data, interpolation="nearest")
        text = "First pixel should be centered on (0, 0) coordinates"
        label = make.label(text, (1, 1), (0, 0), "L")
        rect = make.rectangle(5, 5, 10, 10, "Rectangle")
        cursors = []
        for i_cursor in range(0, 21, 10):
            cursors.append(make.vcursor(i_cursor, movable=False))
            cursors.append(make.hcursor(i_cursor, movable=False))
        win = ptv.show_items([image, label, rect] + cursors, wintitle=title)
        plot = win.get_plot()
        plot.select_item(image)
        win.manager.get_tool(DisplayCoordsTool).activate_curve_pointer(True)


if __name__ == "__main__":
    test_pixel_coords()



# -*- coding: utf-8 -*-
"""
ResizeDialog test
"""

# guitest: show

import pytest
from guidata.qthelpers import exec_dialog, qt_app_context
from qtpy import QtCore as QC

from plotpy.widgets.resizedialog import ResizeDialog

size_list = [  # new_size, old_size, kepp_size, result[widtg, height, zoom]
    ((150, 100), (300, 200), False, (150, 100, 0.5)),
    ((800, 400), (200, 100), False, (800, 400, 4)),
    ((800, 400), (200, 100), True, (800, 400, 1)),
]


@pytest.mark.parametrize("new_size,old_size,keep_original_size,result", size_list)
def test_resize_dialog(new_size, old_size, keep_original_size, result):
    with qt_app_context():
        dialog = ResizeDialog(
            None, new_size, old_size, "Enter the new size:", keep_original_size
        )
        assert result[0] == dialog.width
        assert result[1] == dialog.height
        assert result[2] == dialog.get_zoom()
        assert dialog.keep_original_size is keep_original_size


def test_resize_dialog_qtbot_accept():
    with qt_app_context():
        result = (1500, 1000, 5)
        dialog = ResizeDialog(None, (150, 100), (300, 200), "Enter the new size:")
        dialog.show()
        dialog.w_edit.setText("1500")
        QC.QTimer.singleShot(10, dialog.accept)
        dialog.exec()
        assert result[0] == dialog.width
        assert result[1] == dialog.height
        assert result[2] == dialog.get_zoom()
        assert dialog.keep_original_size is False


def test_resize_dialog_qtbot_reject():
    with qt_app_context():
        result = (150, 100, 0.5)
        dialog = ResizeDialog(None, (150, 100), (300, 200), "Enter the new size:")
        dialog.show()
        exec_dialog(dialog)

        assert result[0] == dialog.width
        assert result[1] == dialog.height
        assert result[2] == dialog.get_zoom()
        assert dialog.keep_original_size is False


def simple_test():
    """ResizeDialog simple test"""
    with qt_app_context():
        dialog = ResizeDialog(None, (150, 100), (300, 250), "Enter the new size:")
        exec_dialog(dialog)


if __name__ == "__main__":
    simple_test()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Widget/plot builder
-------------------

This module provides a set of factory functions to simplify the creation of
plot widgets and grid items.
"""

# Note: when adding method to builder classes, please do not forget to update the
# documentation (see builder.rst file). Because of class inheritance, the methods
# are not automatically documented (otherwise, they would be sorted alphabetically,
# due to a limitation of sphinx auto-doc).

from __future__ import annotations

from typing import TYPE_CHECKING

from plotpy.config import CONF, _
from plotpy.items import GridItem
from plotpy.plot import PlotDialog, PlotOptions, PlotWidget, PlotWindow
from plotpy.styles import COLORS, GridParam, LineStyleParam

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget

    from plotpy.constants import PlotType
    from plotpy.panels import PanelWidget


class WidgetBuilder:
    """Class regrouping a set of factory functions to simplify the creation
    of plot widgets and grid items."""

    # ---- Plot widgets ---------------------------------------------------------------
    def widget(
        self,
        parent: QWidget | None = None,
        toolbar: bool = False,
        panels: tuple[PanelWidget] | None = None,
        auto_tools: bool = True,
        title: str | None = None,
        xlabel: str | tuple[str, str] | None = None,
        ylabel: str | tuple[str, str] | None = None,
        zlabel: str | None = None,
        xunit: str | tuple[str, str] | None = None,
        yunit: str | tuple[str, str] | None = None,
        zunit: str | None = None,
        yreverse: bool | None = None,
        aspect_ratio: float = 1.0,
        lock_aspect_ratio: bool | None = None,
        curve_antialiasing: bool | None = None,
        gridparam: GridParam | None = None,
        section: str = "plot",
        type: str | PlotType = "auto",
        axes_synchronised: bool = False,
        force_colorbar_enabled: bool = False,
        no_image_analysis_widgets: bool = False,
        show_contrast: bool = False,
        show_itemlist: bool = False,
        show_xsection: bool = False,
        show_ysection: bool = False,
        xsection_pos: str = "top",
        ysection_pos: str = "right",
    ) -> PlotWidget:
        """Make a plot widget (:py:class:`.PlotWidget` object)

        Args:
            parent: Parent widget
            toolbar: Show/hide toolbar
            panels: Additionnal panels
            auto_tools: If True, the plot tools are automatically registered.
             If False, the user must register the tools manually.
            title: The plot title
            xlabel: (bottom axis title, top axis title) or bottom axis title only
            ylabel: (left axis title, right axis title) or left axis title only
            zlabel: The Z-axis label
            xunit: (bottom axis unit, top axis unit) or bottom axis unit only
            yunit: (left axis unit, right axis unit) or left axis unit only
            zunit: The Z-axis unit
            yreverse: If True, the Y-axis is reversed
            aspect_ratio: The plot aspect ratio
            lock_aspect_ratio: If True, the aspect ratio is locked
            curve_antialiasing: If True, the curve antialiasing is enabled
            gridparam: The grid parameters
            section: The plot configuration section name ("plot", by default)
            type: The plot type ("auto", "manual", "curve" or "image")
            axes_synchronised: If True, the axes are synchronised
            force_colorbar_enabled: If True, the colorbar is always enabled
            no_image_analysis_widgets: If True, the image analysis widgets are not added
            show_contrast: If True, the contrast adjustment panel is visible
            show_itemlist: If True, the itemlist panel is visible
            show_xsection: If True, the X-axis cross section panel is visible
            show_ysection: If True, the Y-axis cross section panel is visible
            xsection_pos: The X-axis cross section panel position ("top" or "bottom")
            ysection_pos: The Y-axis cross section panel position ("left" or "right")
        """
        options = PlotOptions(
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            zlabel=zlabel,
            xunit=xunit,
            yunit=yunit,
            zunit=zunit,
            yreverse=yreverse,
            aspect_ratio=aspect_ratio,
            lock_aspect_ratio=lock_aspect_ratio,
            curve_antialiasing=curve_antialiasing,
            gridparam=gridparam,
            section=section,
            type=type,
            axes_synchronised=axes_synchronised,
            force_colorbar_enabled=force_colorbar_enabled,
            no_image_analysis_widgets=no_image_analysis_widgets,
            show_contrast=show_contrast,
            show_itemlist=show_itemlist,
            show_xsection=show_xsection,
            show_ysection=show_ysection,
            xsection_pos=xsection_pos,
            ysection_pos=ysection_pos,
        )
        return PlotWidget(
            parent,
            toolbar=toolbar,
            options=options,
            panels=panels,
            auto_tools=auto_tools,
        )

    def dialog(
        self,
        parent: QWidget | None = None,
        toolbar: bool = False,
        panels: tuple[PanelWidget] | None = None,
        auto_tools: bool = True,
        wintitle: str = "PlotPy",
        icon: str = "plotpy.svg",
        edit: bool = False,
        size: tuple[int, int] | None = None,
        title: str | None = None,
        xlabel: str | tuple[str, str] | None = None,
        ylabel: str | tuple[str, str] | None = None,
        zlabel: str | None = None,
        xunit: str | tuple[str, str] | None = None,
        yunit: str | tuple[str, str] | None = None,
        zunit: str | None = None,
        yreverse: bool | None = None,
        aspect_ratio: float = 1.0,
        lock_aspect_ratio: bool | None = None,
        curve_antialiasing: bool | None = None,
        gridparam: GridParam | None = None,
        section: str = "plot",
        type: str | PlotType = "auto",
        axes_synchronised: bool = False,
        force_colorbar_enabled: bool = False,
        no_image_analysis_widgets: bool = False,
        show_contrast: bool = False,
        show_itemlist: bool = False,
        show_xsection: bool = False,
        show_ysection: bool = False,
        xsection_pos: str = "top",
        ysection_pos: str = "right",
    ) -> PlotDialog:
        """Make a plot dialog (:py:class:`.PlotDialog` object)

        Args:
            parent: Parent widget
            toolbar: Show/hide toolbar
            panels: Additionnal panels
            auto_tools: If True, the plot tools are automatically registered.
             If False, the user must register the tools manually.
            wintitle: The window title
            icon: The window icon
            edit: If True, the plot is editable
            size: The window size
            title: The plot title
            xlabel: (bottom axis title, top axis title) or bottom axis title only
            ylabel: (left axis title, right axis title) or left axis title only
            zlabel: The Z-axis label
            xunit: (bottom axis unit, top axis unit) or bottom axis unit only
            yunit: (left axis unit, right axis unit) or left axis unit only
            zunit: The Z-axis unit
            yreverse: If True, the Y-axis is reversed
            aspect_ratio: The plot aspect ratio
            lock_aspect_ratio: If True, the aspect ratio is locked
            curve_antialiasing: If True, the curve antialiasing is enabled
            gridparam: The grid parameters
            section: The plot configuration section name ("plot", by default)
            type: The plot type ("auto", "manual", "curve" or "image")
            axes_synchronised: If True, the axes are synchronised
            force_colorbar_enabled: If True, the colorbar is always enabled
            no_image_analysis_widgets: If True, the image analysis widgets are not added
            show_contrast: If True, the contrast adjustment panel is visible
            show_itemlist: If True, the itemlist panel is visible
            show_xsection: If True, the X-axis cross section panel is visible
            show_ysection: If True, the Y-axis cross section panel is visible
            xsection_pos: The X-axis cross section panel position ("top" or "bottom")
            ysection_pos: The Y-axis cross section panel position ("left" or "right")
        """
        options = PlotOptions(
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            zlabel=zlabel,
            xunit=xunit,
            yunit=yunit,
            zunit=zunit,
            yreverse=yreverse,
            aspect_ratio=aspect_ratio,
            lock_aspect_ratio=lock_aspect_ratio,
            curve_antialiasing=curve_antialiasing,
            gridparam=gridparam,
            section=section,
            type=type,
            axes_synchronised=axes_synchronised,
            force_colorbar_enabled=force_colorbar_enabled,
            no_image_analysis_widgets=no_image_analysis_widgets,
            show_contrast=show_contrast,
            show_itemlist=show_itemlist,
            show_xsection=show_xsection,
            show_ysection=show_ysection,
            xsection_pos=xsection_pos,
            ysection_pos=ysection_pos,
        )
        return PlotDialog(
            parent,
            toolbar=toolbar,
            options=options,
            panels=panels,
            auto_tools=auto_tools,
            title=wintitle,
            icon=icon,
            edit=edit,
            size=size,
        )

    def window(
        self,
        parent: QWidget | None = None,
        toolbar: bool = False,
        panels: tuple[PanelWidget] | None = None,
        auto_tools: bool = True,
        wintitle: str = "PlotPy",
        icon: str = "plotpy.svg",
        size: tuple[int, int] | None = None,
        title: str | None = None,
        xlabel: str | tuple[str, str] | None = None,
        ylabel: str | tuple[str, str] | None = None,
        zlabel: str | None = None,
        xunit: str | tuple[str, str] | None = None,
        yunit: str | tuple[str, str] | None = None,
        zunit: str | None = None,
        yreverse: bool | None = None,
        aspect_ratio: float = 1.0,
        lock_aspect_ratio: bool | None = None,
        curve_antialiasing: bool | None = None,
        gridparam: GridParam | None = None,
        section: str = "plot",
        type: str | PlotType = "auto",
        axes_synchronised: bool = False,
        force_colorbar_enabled: bool = False,
        no_image_analysis_widgets: bool = False,
        show_contrast: bool = False,
        show_itemlist: bool = False,
        show_xsection: bool = False,
        show_ysection: bool = False,
        xsection_pos: str = "top",
        ysection_pos: str = "right",
    ) -> PlotWindow:
        """Make a plot window (:py:class:`.PlotWindow` object)

        Args:
            parent: Parent widget
            toolbar: Show/hide toolbar
            panels: Additionnal panels
            auto_tools: If True, the plot tools are automatically registered.
             If False, the user must register the tools manually.
            wintitle: The window title
            icon: The window icon
            size: The window size
            title: The plot title
            xlabel: (bottom axis title, top axis title) or bottom axis title only
            ylabel: (left axis title, right axis title) or left axis title only
            zlabel: The Z-axis label
            xunit: (bottom axis unit, top axis unit) or bottom axis unit only
            yunit: (left axis unit, right axis unit) or left axis unit only
            zunit: The Z-axis unit
            yreverse: If True, the Y-axis is reversed
            aspect_ratio: The plot aspect ratio
            lock_aspect_ratio: If True, the aspect ratio is locked
            curve_antialiasing: If True, the curve antialiasing is enabled
            gridparam: The grid parameters
            section: The plot configuration section name ("plot", by default)
            type: The plot type ("auto", "manual", "curve" or "image")
            axes_synchronised: If True, the axes are synchronised
            force_colorbar_enabled: If True, the colorbar is always enabled
            no_image_analysis_widgets: If True, the image analysis widgets are not added
            show_contrast: If True, the contrast adjustment panel is visible
            show_itemlist: If True, the itemlist panel is visible
            show_xsection: If True, the X-axis cross section panel is visible
            show_ysection: If True, the Y-axis cross section panel is visible
            xsection_pos: The X-axis cross section panel position ("top" or "bottom")
            ysection_pos: The Y-axis cross section panel position ("left" or "right")
        """
        options = PlotOptions(
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            zlabel=zlabel,
            xunit=xunit,
            yunit=yunit,
            zunit=zunit,
            yreverse=yreverse,
            aspect_ratio=aspect_ratio,
            lock_aspect_ratio=lock_aspect_ratio,
            curve_antialiasing=curve_antialiasing,
            gridparam=gridparam,
            section=section,
            type=type,
            axes_synchronised=axes_synchronised,
            force_colorbar_enabled=force_colorbar_enabled,
            no_image_analysis_widgets=no_image_analysis_widgets,
            show_contrast=show_contrast,
            show_itemlist=show_itemlist,
            show_xsection=show_xsection,
            show_ysection=show_ysection,
            xsection_pos=xsection_pos,
            ysection_pos=ysection_pos,
        )
        return PlotWindow(
            parent,
            toolbar=toolbar,
            options=options,
            panels=panels,
            auto_tools=auto_tools,
            title=wintitle,
            icon=icon,
            size=size,
        )

    def gridparam(
        self,
        background: str | None = None,
        major_enabled: tuple[bool, bool] | None = None,
        minor_enabled: tuple[bool, bool] | None = None,
        major_style: tuple[str, str, int] | None = None,
        minor_style: tuple[str, str, int] | None = None,
    ) -> GridParam:
        """Make :py:class:`.GridParam` instance

        Args:
            background: canvas background color
            major_enabled: major grid enabled (x, y)
            minor_enabled: minor grid enabled (x, y)
            major_style: major grid style (linestyle, color, width)
            minor_style: minor grid style (linestyle, color, width)

        Returns:
            :py:class:`.GridParam`: grid parameters
        """
        gridparam = GridParam(title=_("Grid"), icon="lin_lin.png")
        gridparam.read_config(CONF, "plot", "grid")
        if background is not None:
            gridparam.background = background
        if major_enabled is not None:
            gridparam.maj_xenabled, gridparam.maj_yenabled = major_enabled
        if minor_enabled is not None:
            gridparam.min_xenabled, gridparam.min_yenabled = minor_enabled
        if major_style is not None:
            style = LineStyleParam()
            linestyle, color, style.width = major_style
            style.set_style_from_matlab(linestyle)
            style.color = COLORS.get(color, color)  # MATLAB-style
        if minor_style is not None:
            style = LineStyleParam()
            linestyle, color, style.width = minor_style
            style.set_style_from_matlab(linestyle)
            style.color = COLORS.get(color, color)  # MATLAB-style
        return gridparam

    def grid(
        self,
        background: str | None = None,
        major_enabled: tuple[bool, bool] | None = None,
        minor_enabled: tuple[bool, bool] | None = None,
        major_style: tuple[str, str, int] | None = None,
        minor_style: tuple[str, str, int] | None = None,
    ) -> GridItem:
        """Make a grid `plot item` (:py:class:`.GridItem` object)

        Args:
            background: canvas background color
            major_enabled: major grid enabled (x, y)
            minor_enabled: minor grid enabled (x, y)
            major_style: major grid style (linestyle, color, width)
            minor_style: minor grid style (linestyle, color, width)

        Returns:
            :py:class:`.GridItem`: grid item
        """
        gridparam = self.gridparam(
            background, major_enabled, minor_enabled, major_style, minor_style
        )
        return GridItem(gridparam)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Annotations
-----------

The :mod:`annotation` module provides annotated shape plot items.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import numpy as np
from guidata.configtools import get_icon
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid

from plotpy.config import CONF, _
from plotpy.coords import canvas_to_axes
from plotpy.interfaces import IBasePlotItem, ISerializableType, IShapeItemType
from plotpy.items.label import DataInfoLabel
from plotpy.items.shape.base import AbstractShape
from plotpy.items.shape.ellipse import EllipseShape
from plotpy.items.shape.point import PointShape
from plotpy.items.shape.polygon import PolygonShape
from plotpy.items.shape.rectangle import ObliqueRectangleShape, RectangleShape
from plotpy.items.shape.segment import SegmentShape
from plotpy.mathutils.geometry import (
    compute_angle,
    compute_center,
    compute_distance,
    compute_rect_size,
)
from plotpy.styles.label import LabelParam
from plotpy.styles.shape import AnnotationParam

if TYPE_CHECKING:
    import guidata.io
    import qwt.scale_map
    from qtpy.QtCore import QPointF, QRectF
    from qtpy.QtGui import QPainter

    from plotpy.interfaces import IItemType
    from plotpy.styles.base import ItemParameters


class AnnotatedShape(AbstractShape):
    """
    Construct an annotated shape with properties set with
    *annotationparam* (see :py:class:`.styles.AnnotationParam`)

    Args:
        annotationparam: Annotation parameters
    """

    __implements__ = (IBasePlotItem, ISerializableType)
    SHAPE_CLASS: type[AbstractShape] = RectangleShape  # to be overridden
    LABEL_ANCHOR: str = ""

    def __init__(
        self,
        annotationparam: AnnotationParam | None = None,
        info_callback: Callable[[AnnotatedShape], str] | None = None,
    ) -> None:
        super().__init__()
        assert self.LABEL_ANCHOR is not None and len(self.LABEL_ANCHOR) != 0
        if info_callback is None:

            def info_callback(annotation: AnnotatedShape) -> str:
                """Return information on annotation"""
                return annotation.get_infos()

        self.info_callback = info_callback
        self.shape: AbstractShape = self.create_shape()
        self.label = self.create_label()
        self.area_computations_visible = True
        self.subtitle_visible = True
        if annotationparam is None:
            self.annotationparam = AnnotationParam(
                _("Annotation"), icon="annotation.png"
            )
        else:
            self.annotationparam = annotationparam
            self.annotationparam.update_item(self)
        self.setIcon(get_icon("annotation.png"))

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IShapeItemType, ISerializableType)

    def __reduce__(self) -> tuple[type, tuple, tuple]:
        """Return a tuple for pickling"""
        self.annotationparam.update_param(self)
        state = (self.shape, self.label, self.annotationparam)
        return (self.__class__, (), state)

    def __setstate__(self, state: tuple) -> None:
        """Set state after unpickling"""
        shape, label, param = state
        self.shape = shape
        self.label = label
        self.annotationparam: AnnotationParam = param
        self.annotationparam.update_item(self)

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        writer.write(self.annotationparam, group_name="annotationparam")
        self.shape.serialize(writer)
        self.label.serialize(writer)

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        self.annotationparam = AnnotationParam(_("Annotation"), icon="annotation.png")
        reader.read("annotationparam", instance=self.annotationparam)
        self.annotationparam.update_item(self)
        self.shape.deserialize(reader)
        self.label.deserialize(reader)

    def set_style(self, section: str, option: str) -> None:
        """Set style for this item

        Args:
            section: Section
            option: Option
        """
        self.shape.set_style(section, option)

    # ----QwtPlotItem API--------------------------------------------------------
    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        self.shape.draw(painter, xMap, yMap, canvasRect)
        if self.label.isVisible():
            self.label.draw(painter, xMap, yMap, canvasRect)

    # ----Public API-------------------------------------------------------------
    def create_shape(self):
        """Return the shape object associated to this annotated shape object"""
        shape = self.SHAPE_CLASS(0, 0, 1, 1)  # pylint: disable=not-callable
        return shape

    def create_label(self) -> DataInfoLabel:
        """Return the label object associated to this annotated shape object

        Returns:
            Label object
        """
        label_param = LabelParam(_("Label"), icon="label.png")
        label_param.read_config(CONF, "plot", "shape/label")
        label_param.anchor = self.LABEL_ANCHOR
        return DataInfoLabel(label_param, [self])

    def is_label_visible(self) -> bool:
        """Return True if associated label is visible

        Returns:
            True if associated label is visible
        """
        return self.label.isVisible()

    def set_label_visible(self, state: bool) -> None:
        """Set the annotated shape's label visibility

        Args:
            state: True if label should be visible
        """
        self.label.setVisible(state)

    def update_label(self) -> None:
        """Update the annotated shape's label contents"""
        self.label.update_text()

    def set_info_callback(self, callback: Callable[[AnnotatedShape], str]) -> None:
        """Set the callback function to get informations on current shape

        Args:
            callback: Callback function to get informations on current shape
        """
        self.info_callback = callback

    def get_text(self) -> str:
        """Return text associated to current shape
        (see :py:class:`.label.ObjectInfo`)

        Returns:
            Text associated to current shape
        """
        text = ""
        title = self.title().text()
        if title:
            text += f"<b>{title}</b>"
        subtitle = self.annotationparam.subtitle
        if subtitle and self.subtitle_visible:
            if text:
                text += "<br>"
            text += f"<i>{subtitle}</i>"
        if self.area_computations_visible:
            infos = self.info_callback(self)
            if infos:
                if text:
                    text += "<br>"
                text += infos
        return text

    def x_to_str(self, x: float) -> str:
        """Convert x to a string (with associated unit and uncertainty)

        Args:
            x: X value

        Returns:
            str: Formatted string with x value
        """
        param = self.annotationparam
        if self.plot() is None:
            return ""
        else:
            xunit = self.plot().get_axis_unit(self.xAxis())
            fmt = param.format
            if param.uncertainty:
                fmt += " ± " + (fmt % (x * param.uncertainty))
            if xunit is not None:
                return (fmt + " " + xunit) % x
            else:
                return (fmt) % x

    def y_to_str(self, y):
        """Convert y to a string (with associated unit and uncertainty)

        Args:
            y: Y value

        Returns:
            str: Formatted string with x value
        """
        param = self.annotationparam
        if self.plot() is None:
            return ""
        else:
            yunit = self.plot().get_axis_unit(self.yAxis())
            fmt = param.format
            if param.uncertainty:
                fmt += " ± " + (fmt % (y * param.uncertainty))
            if yunit is not None:
                return (fmt + " " + yunit) % y
            else:
                return (fmt) % y

    def get_center(self):
        """Return shape center coordinates: (xc, yc)"""
        return self.shape.get_center()

    def get_tr_center(self):
        """Return shape center coordinates after applying transform matrix"""
        raise NotImplementedError

    def get_tr_center_str(self):
        """Return center coordinates as a string (with units)"""
        xc, yc = self.get_tr_center()
        return f"( {self.x_to_str(xc)} ; {self.y_to_str(yc)} )"

    def get_tr_size(self):
        """Return shape size after applying transform matrix"""
        raise NotImplementedError

    def get_tr_size_str(self):
        """Return size as a string (with units)"""
        xs, ys = self.get_tr_size()
        return f"{self.x_to_str(xs)} x {self.y_to_str(ys)}"

    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        return ""

    def set_label_position(self):
        """Set label position, for instance based on shape position"""
        raise NotImplementedError

    def apply_transform_matrix(self, x, y):
        """

        :param x:
        :param y:
        :return:
        """
        V = np.array([x, y, 1.0])
        W = np.dot(V, self.annotationparam.transform_matrix)
        return W[0], W[1]

    def get_transformed_coords(self, handle1, handle2):
        """

        :param handle1:
        :param handle2:
        :return:
        """
        x1, y1 = self.apply_transform_matrix(*self.shape.points[handle1])
        x2, y2 = self.apply_transform_matrix(*self.shape.points[handle2])
        return x1, y1, x2, y2

    # ----IBasePlotItem API------------------------------------------------------
    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        return self.shape.poly_hit_test(self.plot(), self.xAxis(), self.yAxis(), pos)

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        self.shape.move_point_to(handle, pos, ctrl)
        self.set_label_position()
        if self.plot():
            self.plot().SIG_ANNOTATION_CHANGED.emit(self)

    def move_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        self.shape.move_shape(old_pos, new_pos)
        self.label.move_local_shape(old_pos, new_pos)

    def move_local_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in canvas coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        old_pt = canvas_to_axes(self, old_pos)
        new_pt = canvas_to_axes(self, new_pos)
        self.shape.move_shape(old_pt, new_pt)
        self.set_label_position()
        if self.plot():
            self.plot().SIG_ITEM_MOVED.emit(self, *(old_pt + new_pt))
            self.plot().SIG_ANNOTATION_CHANGED.emit(self)

    def move_with_selection(self, delta_x: float, delta_y: float) -> None:
        """Translate the item together with other selected items

        Args:
            delta_x: Translation in plot coordinates along x-axis
            delta_y: Translation in plot coordinates along y-axis
        """
        self.shape.move_with_selection(delta_x, delta_y)
        self.label.move_with_selection(delta_x, delta_y)
        self.plot().SIG_ANNOTATION_CHANGED.emit(self)

    def select(self) -> None:
        """
        Select the object and eventually change its appearance to highlight the
        fact that it's selected
        """
        AbstractShape.select(self)
        self.shape.select()

    def unselect(self) -> None:
        """
        Unselect the object and eventually restore its original appearance to
        highlight the fact that it's not selected anymore
        """
        AbstractShape.unselect(self)
        self.shape.unselect()

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        self.shape.get_item_parameters(itemparams)
        self.label.get_item_parameters(itemparams)
        self.annotationparam.update_param(self)
        itemparams.add("AnnotationParam", self, self.annotationparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        self.shape.set_item_parameters(itemparams)
        self.label.set_item_parameters(itemparams)
        update_dataset(
            self.annotationparam, itemparams.get("AnnotationParam"), visible_only=True
        )
        self.annotationparam.update_item(self)
        self.plot().SIG_ANNOTATION_CHANGED.emit(self)

    # Autoscalable types API

    def is_empty(self) -> bool:
        """Return True if the item is empty

        Returns:
            True if the item is empty, False otherwise
        """
        return self.shape.is_empty()

    def boundingRect(self) -> QRectF:
        """Return the bounding rectangle of the shape

        Returns:
            Bounding rectangle of the shape
        """
        return self.shape.boundingRect()


assert_interfaces_valid(AnnotatedShape)


class AnnotatedPoint(AnnotatedShape):
    """
    Construct an annotated point at coordinates (x, y) with properties set with
    *annotationparam* (see :py:class:`.styles.AnnotationParam`)
    """

    SHAPE_CLASS = PointShape
    LABEL_ANCHOR = "TL"

    def __init__(
        self,
        x=0,
        y=0,
        annotationparam: AnnotationParam | None = None,
        info_callback: Callable[[AnnotatedShape], str] | None = None,
    ) -> None:
        super().__init__(annotationparam, info_callback)
        self.shape: PointShape
        self.set_pos(x, y)
        self.setIcon(get_icon("point_shape.png"))

    # ----Public API-------------------------------------------------------------
    def set_pos(self, x, y):
        """Set the point coordinates to (x, y)"""
        self.shape.set_pos(x, y)
        self.set_label_position()

    def get_pos(self):
        """Return the point coordinates"""
        return self.shape.get_pos()

    # ----AnnotatedShape API-----------------------------------------------------
    def create_shape(self):
        """Return the shape object associated to this annotated shape object"""
        shape = self.SHAPE_CLASS(0, 0)
        return shape

    def set_label_position(self):
        """Set label position, for instance based on shape position"""
        x, y = self.shape.points[0]
        self.label.set_pos(x, y)

    # ----AnnotatedShape API-----------------------------------------------------
    def get_tr_position(self):
        xt, yt = self.apply_transform_matrix(*self.shape.points[0])
        return xt, yt

    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        xt, yt = self.apply_transform_matrix(*self.shape.points[0])
        s = "{title} ( {posx} ; {posy} )"
        s = s.format(
            title=_("Position:"), posx=self.x_to_str(xt), posy=self.y_to_str(yt)
        )
        return s


class AnnotatedSegment(AnnotatedShape):
    """
    Construct an annotated segment between coordinates (x1, y1) and
    (x2, y2) with properties set with *annotationparam*
    (see :py:class:`.styles.AnnotationParam`)
    """

    SHAPE_CLASS = SegmentShape
    LABEL_ANCHOR = "C"

    def __init__(
        self,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        annotationparam: AnnotationParam | None = None,
        info_callback: Callable[[AnnotatedShape], str] | None = None,
    ) -> None:
        super().__init__(annotationparam, info_callback)
        self.shape: SegmentShape
        self.set_rect(x1, y1, x2, y2)
        self.setIcon(get_icon("segment.png"))

    # ----Public API-------------------------------------------------------------
    def set_rect(self, x1, y1, x2, y2):
        """
        Set the coordinates of the shape's top-left corner to (x1, y1),
        and of its bottom-right corner to (x2, y2).
        """
        self.shape.set_rect(x1, y1, x2, y2)
        self.set_label_position()

    def get_rect(self):
        """
        Return the coordinates of the shape's top-left and bottom-right corners
        """
        return self.shape.get_rect()

    def get_tr_length(self):
        """Return segment length after applying transform matrix"""
        return compute_distance(*self.get_transformed_coords(0, 1))

    def get_tr_center(self):
        """Return segment position (middle) after applying transform matrix"""
        return compute_center(*self.get_transformed_coords(0, 1))

    # ----AnnotatedShape API-----------------------------------------------------
    def set_label_position(self):
        """Set label position, for instance based on shape position"""
        x1, y1, x2, y2 = self.get_rect()
        self.label.set_pos(*compute_center(x1, y1, x2, y2))

    # ----AnnotatedShape API-----------------------------------------------------
    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        return "<br>".join(
            [
                _("Center:") + " " + self.get_tr_center_str(),
                _("Distance:") + " " + self.x_to_str(self.get_tr_length()),
            ]
        )


class AnnotatedPolygon(AnnotatedShape):
    """
    Construct an annotated polygon with properties set with *annotationparam*
    (see :py:class:`.styles.AnnotationParam`)

    Args:
        points: List of points
        closed: True if polygon is closed
        annotationparam: Annotation parameters
    """

    SHAPE_CLASS = PolygonShape
    LABEL_ANCHOR = "C"

    def __init__(
        self,
        points: list[tuple[float, float]] | None = None,
        closed: bool | None = None,
        annotationparam: AnnotationParam | None = None,
        info_callback: Callable[[AnnotatedShape], str] | None = None,
    ) -> None:
        super().__init__(annotationparam, info_callback)
        self.shape: PolygonShape
        if points is not None:
            self.set_points(points)
        if closed is not None:
            self.set_closed(closed)
        self.setIcon(get_icon("polygon.png"))

    # ----Public API-------------------------------------------------------------
    def set_points(self, points: list[tuple[float, float]] | np.ndarray | None) -> None:
        """Set the polygon points

        Args:
            points: List of point coordinates
        """
        self.shape.set_points(points)
        self.set_label_position()

    def get_points(self) -> np.ndarray:
        """Return polygon points

        Returns:
            Polygon points (array of shape (N, 2))
        """
        return self.shape.get_points()

    def set_closed(self, state: bool) -> None:
        """Set closed state

        Args:
            state: True if the polygon is closed, False otherwise
        """
        self.shape.set_closed(state)

    def is_closed(self) -> bool:
        """Return True if the polygon is closed, False otherwise

        Returns:
            True if the polygon is closed, False otherwise
        """
        return self.shape.is_closed()

    def is_empty(self) -> bool:
        """Return True if the item is empty

        Returns:
            True if the item is empty, False otherwise
        """
        return self.shape.is_empty()

    def add_local_point(self, pos: tuple[float, float]) -> int:
        """Add a point in canvas coordinates (local coordinates)

        Args:
            pos: Position

        Returns:
            Handle of the added point
        """
        pt = canvas_to_axes(self, pos)
        return self.add_point(pt)

    def add_point(self, pt: tuple[float, float]) -> int:
        """Add a point in axis coordinates

        Args:
            pt: Position

        Returns:
            Handle of the added point
        """
        handle = self.shape.add_point(pt)
        self.set_label_position()
        return handle

    def del_point(self, handle: int) -> int:
        """Delete a point

        Args:
            handle: Handle

        Returns:
            Handle of the deleted point
        """
        handle = self.shape.del_point(handle)
        self.set_label_position()
        return handle

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        pt = canvas_to_axes(self, pos)
        self.move_point_to(handle, pt)

    def move_shape(
        self, old_pos: tuple[float, float], new_pos: tuple[float, float]
    ) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        self.shape.move_shape(old_pos, new_pos)
        self.set_label_position()

    # ----AnnotatedShape API-----------------------------------------------------
    def create_shape(self):
        """Return the shape object associated to this annotated shape object"""
        shape = self.SHAPE_CLASS()  # pylint: disable=not-callable
        return shape

    def set_label_position(self) -> None:
        """Set label position, for instance based on shape position"""
        x, y = self.shape.get_center()
        self.label.set_pos(x, y)

    def get_tr_center(self) -> tuple[float, float]:
        """Return shape center coordinates after applying transform matrix"""
        return self.shape.get_center()

    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        return "<br>".join(
            [
                _("Center:") + " " + self.get_tr_center_str(),
            ]
        )


class AnnotatedRectangle(AnnotatedShape):
    """
    Construct an annotated rectangle between coordinates (x1, y1) and
    (x2, y2) with properties set with *annotationparam*
    (see :py:class:`.styles.AnnotationParam`)
    """

    SHAPE_CLASS = RectangleShape
    LABEL_ANCHOR = "TL"

    def __init__(
        self,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        annotationparam: AnnotationParam | None = None,
        info_callback: Callable[[AnnotatedShape], str] | None = None,
    ) -> None:
        super().__init__(annotationparam, info_callback)
        self.shape: RectangleShape
        self.set_rect(x1, y1, x2, y2)
        self.setIcon(get_icon("rectangle.png"))

    # ----Public API-------------------------------------------------------------
    def set_rect(self, x1, y1, x2, y2):
        """
        Set the coordinates of the shape's top-left corner to (x1, y1),
        and of its bottom-right corner to (x2, y2).
        """
        self.shape.set_rect(x1, y1, x2, y2)
        self.set_label_position()

    def get_rect(self) -> tuple[float, float, float, float]:
        """
        Return the coordinates of the shape's top-left and bottom-right corners
        """
        return self.shape.get_rect()

    # ----AnnotatedShape API-----------------------------------------------------
    def set_label_position(self):
        """Set label position, for instance based on shape position"""
        x_label, y_label = self.shape.points.min(axis=0)
        self.label.set_pos(x_label, y_label)

    def get_tr_center(self):
        """Return shape center coordinates after applying transform matrix"""
        return compute_center(*self.get_transformed_coords(0, 2))

    def get_tr_size(self):
        """Return shape size after applying transform matrix"""
        return compute_rect_size(*self.get_transformed_coords(0, 2))

    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        return "<br>".join(
            [
                _("Center:") + " " + self.get_tr_center_str(),
                _("Size:") + " " + self.get_tr_size_str(),
            ]
        )


class AnnotatedObliqueRectangle(AnnotatedRectangle):
    """
    Construct an annotated oblique rectangle between coordinates (x0, y0),
    (x1, y1), (x2, y2) and (x3, y3) with properties set with *annotationparam*
    (see :py:class:`.styles.AnnotationParam`)
    """

    SHAPE_CLASS = ObliqueRectangleShape
    LABEL_ANCHOR = "C"

    def __init__(
        self, x0=0, y0=0, x1=0, y1=0, x2=0, y2=0, x3=0, y3=0, annotationparam=None
    ):
        AnnotatedShape.__init__(self, annotationparam)
        self.shape: ObliqueRectangleShape
        self.set_rect(x0, y0, x1, y1, x2, y2, x3, y3)
        self.setIcon(get_icon("oblique_rectangle.png"))

    # ----Public API-------------------------------------------------------------
    def get_tr_angle(self):
        """Return X-diameter angle with horizontal direction,
        after applying transform matrix"""
        xcoords = self.get_transformed_coords(0, 1)
        _x, yr1 = self.apply_transform_matrix(1.0, 1.0)
        _x, yr2 = self.apply_transform_matrix(1.0, 2.0)
        return (compute_angle(*xcoords, reverse=yr1 > yr2) + 90) % 180 - 90

    def get_bounding_rect_coords(self) -> tuple[float, float, float, float]:
        """Return bounding rectangle coordinates (in plot coordinates)

        Returns:
            Bounding rectangle coordinates (in plot coordinates)
        """
        return self.shape.get_bounding_rect_coords()

    # ----AnnotatedShape API-----------------------------------------------------
    def create_shape(self):
        """Return the shape object associated to this annotated shape object"""
        shape = self.SHAPE_CLASS(0, 0, 0, 0, 0, 0, 0, 0)
        return shape

    # ----AnnotatedShape API-----------------------------------------------------
    def set_label_position(self):
        """Set label position, for instance based on shape position"""
        self.label.set_pos(*self.get_center())

    # ----RectangleShape API-----------------------------------------------------
    def set_rect(self, x0, y0, x1, y1, x2, y2, x3, y3):
        """
        Set the rectangle corners coordinates:

            (x0, y0): top-left corner
            (x1, y1): top-right corner
            (x2, y2): bottom-right corner
            (x3, y3): bottom-left corner

        ::

            x: additionnal points

            (x0, y0)------>(x1, y1)
                ↑             |
                |             |
                x             x
                |             |
                |             ↓
            (x3, y3)<------(x2, y2)
        """
        self.shape.set_rect(x0, y0, x1, y1, x2, y2, x3, y3)
        self.set_label_position()

    def get_tr_size(self):
        """Return shape size after applying transform matrix"""
        dx = compute_distance(*self.get_transformed_coords(0, 1))
        dy = compute_distance(*self.get_transformed_coords(0, 3))
        return dx, dy

    # ----AnnotatedShape API-----------------------------------------------------
    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        return "<br>".join(
            [
                _("Center:") + " " + self.get_tr_center_str(),
                _("Size:") + " " + self.get_tr_size_str(),
                _("Angle:") + " %.1f°" % self.get_tr_angle(),
            ]
        )

    def get_tr_center(self):
        x0, y0, x2, y2 = self.get_transformed_coords(0, 2)
        return compute_center(x0, y0, x2, y2)


class AnnotatedEllipse(AnnotatedShape):
    """
    Construct an annotated ellipse with X-axis diameter between
    coordinates (x1, y1) and (x2, y2)
    and properties set with *annotationparam*
    (see :py:class:`.styles.AnnotationParam`)
    """

    SHAPE_CLASS = EllipseShape
    LABEL_ANCHOR = "C"

    def __init__(
        self,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        annotationparam: AnnotationParam | None = None,
        info_callback: Callable[[AnnotatedShape], str] | None = None,
    ) -> None:
        super().__init__(annotationparam, info_callback)
        self.shape: EllipseShape
        self.set_xdiameter(x1, y1, x2, y2)
        self.setIcon(get_icon("ellipse_shape.png"))
        self.switch_to_ellipse()

    # ----Public API-------------------------------------------------------------
    def switch_to_ellipse(self):
        """Switch to ellipse mode"""
        self.shape.switch_to_ellipse()

    def switch_to_circle(self):
        """Switch to circle mode"""
        self.shape.switch_to_circle()

    def set_xdiameter(self, x0, y0, x1, y1):
        """Set the coordinates of the ellipse's X-axis diameter
        Warning: transform matrix is not applied here"""
        self.shape.set_xdiameter(x0, y0, x1, y1)
        self.set_label_position()

    def get_xdiameter(self):
        """Return the coordinates of the ellipse's X-axis diameter
        Warning: transform matrix is not applied here"""
        return self.shape.get_xdiameter()

    def set_ydiameter(self, x2, y2, x3, y3):
        """Set the coordinates of the ellipse's Y-axis diameter
        Warning: transform matrix is not applied here"""
        self.shape.set_ydiameter(x2, y2, x3, y3)
        self.set_label_position()

    def get_ydiameter(self):
        """Return the coordinates of the ellipse's Y-axis diameter
        Warning: transform matrix is not applied here"""
        return self.shape.get_ydiameter()

    def get_rect(self):
        """

        :return:
        """
        return self.shape.get_rect()

    def set_rect(self, x0, y0, x1, y1):
        """

        :param x0:
        :param y0:
        :param x1:
        :param y1:
        """
        raise NotImplementedError

    def get_tr_angle(self):
        """Return X-diameter angle with horizontal direction,
        after applying transform matrix"""
        xcoords = self.get_transformed_coords(0, 1)
        _x, yr1 = self.apply_transform_matrix(1.0, 1.0)
        _x, yr2 = self.apply_transform_matrix(1.0, 2.0)
        return (compute_angle(*xcoords, reverse=yr1 > yr2) + 90) % 180 - 90

    # ----AnnotatedShape API-----------------------------------------------------
    def set_label_position(self):
        """Set label position, for instance based on shape position"""
        x_label, y_label = self.shape.points.mean(axis=0)
        self.label.set_pos(x_label, y_label)

    def get_tr_center(self):
        """Return center coordinates: (xc, yc)"""
        return compute_center(*self.get_transformed_coords(0, 1))

    def get_tr_size(self):
        """Return shape size after applying transform matrix"""
        xcoords = self.get_transformed_coords(0, 1)
        ycoords = self.get_transformed_coords(2, 3)
        dx = compute_distance(*xcoords)
        dy = compute_distance(*ycoords)
        if np.fabs(self.get_tr_angle()) > 45:
            dx, dy = dy, dx
        return dx, dy

    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        return "<br>".join(
            [
                _("Center:") + " " + self.get_tr_center_str(),
                _("Size:") + " " + self.get_tr_size_str(),
                _("Angle:") + f" {self.get_tr_angle():.1f}°",
            ]
        )


class AnnotatedCircle(AnnotatedEllipse):
    """
    Construct an annotated circle with diameter between coordinates
    (x1, y1) and (x2, y2) and properties set with *annotationparam*
    (see :py:class:`.styles.AnnotationParam`)
    """

    def __init__(self, x1=0, y1=0, x2=0, y2=0, annotationparam=None):
        AnnotatedEllipse.__init__(self, x1, y1, x2, y2, annotationparam)
        self.shape.switch_to_circle()

    def get_tr_diameter(self):
        """Return circle diameter after applying transform matrix"""
        return compute_distance(*self.get_transformed_coords(0, 1))

    # ----AnnotatedShape API-------------------------------------------------
    def get_infos(self) -> str:
        """Get informations on current shape

        Returns:
            str: Formatted string with informations on current shape
        """
        return "<br>".join(
            [
                _("Center:") + " " + self.get_tr_center_str(),
                _("Diameter:") + " " + self.x_to_str(self.get_tr_diameter()),
            ]
        )

    # ----AnnotatedEllipse API---------------------------------------------------
    def set_rect(self, x0, y0, x1, y1):
        """

        :param x0:
        :param y0:
        :param x1:
        :param y1:
        """
        self.shape.set_rect(x0, y0, x1, y1)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Constants
=========

This module provides constants used in plotpy.
"""

from __future__ import annotations

import enum

from plotpy.config import _

# ===============================================================================
# Plot axes
# ===============================================================================

#: ID of the x axis
AXIS_IDS = Y_LEFT, Y_RIGHT, X_BOTTOM, X_TOP = range(4)


# ===============================================================================
# Plot types
# ===============================================================================


class PlotType(enum.Enum):
    """
    This is the enum used for the plot type. Defines how the plot should deal with the
    different PlotItems types (curves and images)
    """

    #: Automatic plot type. The first PlotItem attached to the plot sets the type.
    #: All tools (curve and image related) are registered and accessible depending
    #: on the last selected PlotItem.
    AUTO = 1

    #: Curve specialized plot. The y axis is not reversed and the aspect ratio is not
    #: locked by default. Only CURVE typed tools are automatically registered.
    CURVE = 2

    #: Image specialized plot. The y axis is reversed and the aspect ratio is locked
    #: by default. Only IMAGE typed tools are automatically registered.
    IMAGE = 3

    #: No assumption is made on the type of items to be displayed on the plot. Acts
    #: like the CURVE value of the enum for y axis and aspect ratio. No tool are
    #: automatically registered.
    MANUAL = 4


# ===============================================================================
# Plot parameters
# ===============================================================================

PARAMETERS_TITLE_ICON = {
    "grid": (_("Grid..."), "grid.png"),
    "axes": (_("Axes style..."), "axes.png"),
    "item": (_("Parameters..."), "settings.png"),
}


# ===============================================================================
# Panels
# ===============================================================================

#: ID of the `item list` panel
ID_ITEMLIST = "itemlist"
#: ID of the `contrast adjustment` panel
ID_CONTRAST = "contrast"
#: ID of the `X-axis cross section` panel
ID_XCS = "x_cross_section"
#: ID of the `Y-axis cross section` panel
ID_YCS = "y_cross_section"
#: ID of the `oblique averaged cross section` panel
ID_OCS = "oblique_cross_section"
#: ID of the `line cross section` panel
ID_LCS = "line_cross_section"


# ===============================================================================
# Plot items
# ===============================================================================

# Shape Z offset used when adding shapes to the plot
SHAPE_Z_OFFSET = 1000


# Lookup table alpha functions for image items
class LUTAlpha(enum.Enum):
    """LUT Alpha functions"""

    #: No alpha function
    NONE = 0

    #: Constant alpha function
    CONSTANT = 1

    #: Linear alpha function
    LINEAR = 2

    #: Sigmoid alpha function
    SIGMOID = 3

    #: Hyperbolic tangent alpha function
    TANH = 4

    #: Transparent lowest value and opaque highest value
    STEP = 5

    @classmethod
    def get_member_from_name(cls: type["LUTAlpha"], name: str) -> "LUTAlpha|None":
        """Return the member of the enum from its name"""
        for member in cls:
            if member.name.lower() == name.lower():
                return member
        return None

    @classmethod
    def get_choices(cls: type["LUTAlpha"]) -> list[tuple[int, str]]:
        """Return the list of choices"""
        return [
            (LUTAlpha.NONE.value, _("None")),
            (LUTAlpha.CONSTANT.value, _("Constant")),
            (LUTAlpha.LINEAR.value, _("Linear")),
            (LUTAlpha.SIGMOID.value, _("Sigmoid")),
            (LUTAlpha.TANH.value, _("Hyperbolic tangent")),
            (LUTAlpha.STEP.value, _("Step")),
        ]

    @classmethod
    def get_help(cls: type["LUTAlpha"]) -> str:
        """Return tooltip help for all alpha functions"""
        return _(
            "Alpha function applied to the Look-Up Table (LUT)<br>"
            "to control the transparency of the image:<br>"
            "(maximum opacity is given by the <b>Global alpha</b> parameter)<br><br>"
            "<b>None</b>: No alpha function<br>"
            "<b>Constant</b>: Constant alpha function<br>"
            "<b>Linear</b>: Linear alpha function<br>"
            "<b>Sigmoid</b>: Sigmoid alpha function<br>"
            "<b>Hyperbolic tangent</b>: Hyperbolic tangent alpha function<br>"
            "<b>Step</b>: Lowest value of the LUT is 100% transparent, other "
            "values are opaque"
        )


# Lookup table size
LUT_SIZE = 1024
LUT_MAX = float(LUT_SIZE - 1)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
LUT range
---------

The :mod:`plotpy.lutrange` module provides functions to compute and manipulate
the LUT range of plot items.

.. autofunction:: hist_range_threshold

.. autofunction:: lut_range_threshold
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from plotpy.items import BaseImageItem, Histogram2DItem
    from plotpy.items.histogram import HistDataSource


def hist_range_threshold(
    hist: np.ndarray, bin_edges: np.ndarray, percent: float
) -> tuple[float, float]:
    """Return the range corresponding to the given percent of the histogram

    Args:
        hist (numpy.ndarray): The histogram
        bin_edges (numpy.ndarray): The bin edges
        percent (float): The percent of the histogram

    Returns:
        tuple[float, float]: The range
    """
    hist = np.concatenate((hist, [0]))
    hist = hist[1:]
    bin_edges = bin_edges[1:]
    threshold = 0.5 * percent / 100 * hist.sum()

    i_bin_min = np.cumsum(hist).searchsorted(threshold)
    i_bin_max = -1 - np.cumsum(np.flipud(hist)).searchsorted(threshold)

    return bin_edges[i_bin_min], bin_edges[i_bin_max]


def lut_range_threshold(
    item: BaseImageItem | Histogram2DItem | HistDataSource, bins: int, percent: float
) -> tuple[float, float]:
    """Return the lut range corresponding to the given percent of the histogram

    Args:
        item (BaseImageItem | Histogram2DItem | HistDataSource): The plot item
        bins (int): The number of bins
        percent (float): The percent of the histogram

    Returns:
        tuple[float, float]: The lut range
    """
    hist, bin_edges = item.get_histogram(bins)
    return hist_range_threshold(hist, bin_edges, percent)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Tests around image transforms: rotation, translation, ..."""

# guitest: show

from __future__ import annotations

import os

import numpy as np
import pytest
from guidata.env import execenv
from guidata.qthelpers import qt_app_context
from qtpy import QtCore as QC
from qtpy import QtGui as QG

from plotpy import io
from plotpy.builder import make
from plotpy.constants import LUTAlpha
from plotpy.items import TrImageItem, assemble_imageitems
from plotpy.tests import vistools as ptv
from plotpy.tests.data import gen_image4

DEFAULT_CHARS = "".join([chr(c) for c in range(32, 256)])


def get_font_array(sz: int, chars: str = DEFAULT_CHARS) -> np.ndarray | None:
    """Return array of font characters

    Args:
        sz: Font size
        chars: Characters to include (default: all printable characters)

    Returns:
        Array of font characters
    """
    font = QG.QFont()
    font.setFixedPitch(True)
    font.setPixelSize(sz)
    font.setStyleStrategy(QG.QFont.NoAntialias)
    dummy = QG.QImage(10, 10, QG.QImage.Format_ARGB32)
    pnt = QG.QPainter(dummy)
    pnt.setFont(font)
    metric = pnt.fontMetrics()
    rct = metric.boundingRect(chars)
    pnt.end()
    h, w = rct.height(), rct.width()
    img = QG.QImage(w, h, QG.QImage.Format_ARGB32)
    paint = QG.QPainter()
    paint.begin(img)
    paint.setFont(font)
    paint.setBrush(QG.QColor(255, 255, 255))
    paint.setPen(QG.QColor(255, 255, 255))
    paint.drawRect(0, 0, w + 1, h + 1)
    paint.setPen(QG.QColor(0, 0, 0))
    paint.setBrush(QG.QColor(0, 0, 0))
    paint.drawText(0, paint.fontMetrics().ascent(), chars)
    paint.end()
    try:
        data = img.bits().asstring(h * w * 4)
    except AttributeError:
        data = img.bits()
    npy: np.ndarray = np.frombuffer(data, np.uint8)
    return npy.reshape(h, w, 4)[:, :, 0]


def write_text_on_array(
    data: np.ndarray,
    x: int,
    y: int,
    sz: int,
    txt: str,
    range: tuple[int, int] | None = None,
) -> None:
    """Write text in image (in-place)

    Args:
        data: Image data
        x: X-coordinate of top-left corner
        y: Y-coordinate of top-left corner
        sz: Font size
        txt: Text to write
        range: Range of values to map to 0-255 (default: None)
    """
    arr = get_font_array(sz, txt)
    if arr is None:
        return
    if range is None:
        m, M = data.min(), data.max()
    else:
        m, M = range
    z = (float(M) - float(m)) * np.array(arr, float) / 255.0 + m
    arr = np.array(z, data.dtype)
    dy, dx = arr.shape
    data[y : y + dy, x : x + dx] = arr


def make_items(N: int) -> list[TrImageItem]:
    """Make test TrImageItem items

    Args:
        N: Image size (N x N)

    Returns:
        List of image items
    """
    data = gen_image4(N, N)
    m = data.min()
    M = data.max()
    items = [make.trimage(data, alpha_function=LUTAlpha.LINEAR, colormap="jet")]
    for dtype in (np.uint8, np.uint16, np.int8, np.int16):
        info = np.iinfo(dtype().dtype)  # pylint: disable=no-value-for-parameter
        s = float((info.max - info.min))
        a1 = s * (data - m) / (M - m)
        img = np.array(a1 + info.min, dtype)
        write_text_on_array(img, 0, 0, int(N / 15.0), dtype.__name__)
        items.append(make.trimage(img, colormap="jet"))
    nc = int(np.sqrt(len(items)) + 1.0)
    maxy, x, y = 0, 0, 0
    w = None
    for index, item in enumerate(items):
        h = item.boundingRect().height()
        if index % nc == 0:
            x = 0
            y += maxy
            maxy = h
        else:
            x += w
            maxy = max(maxy, h)
        w = item.boundingRect().width()
        item.set_transform(x, y, 0.0)
    return items


def save_image(name: str, data: np.ndarray) -> None:
    """Save image to file

    Args:
        name: Base name of file
        data: Image data
    """
    for fname in (name + ".u16.tif", name + ".u8.png"):
        if os.path.exists(fname):
            os.remove(fname)
    size = int(data.nbytes / 1024.0)
    print(f"Saving image: {data.shape[0]} x {data.shape[1]} ({size} KB):")
    print(" --> uint16")
    io.imwrite(name + ".u16.tif", data, dtype=np.uint16, max_range=True)
    print(" --> uint8")
    io.imwrite(name + ".u8.png", data, dtype=np.uint8, max_range=True)


def get_bbox(items: list[TrImageItem]) -> QC.QRectF:
    """Get bounding box of items

    Args:
        items: List of image items

    Returns:
        Bounding box of items
    """
    rectf = QC.QRectF()
    for item in items:
        rectf = rectf.united(item.boundingRect())
    return rectf


def build_image(items: list[TrImageItem]) -> None:
    """Build image from items

    Args:
        items: List of image items
    """
    r = get_bbox(items)
    _x, _y, w, h = r.getRect()
    print("-" * 80)
    print(f"Assemble test1: {int(w)} x {int(h)}")
    dest = assemble_imageitems(items, r, w, h)
    if not execenv.unattended:
        save_image("test1", dest)
    print("-" * 80)
    print(f"Assemble test1: {int(w/4)} x {int(h/4)}")
    dest = assemble_imageitems(items, r, w / 4, h / 4)
    if not execenv.unattended:
        save_image("test2", dest)
    print("-" * 80)


@pytest.mark.parametrize("N", [500])
@pytest.mark.parametrize("assemble_images", [False, True])
def test_transform(N: int, assemble_images: bool) -> None:
    """Test image transforms

    Args:
        N: Image size (N x N)
        assemble_images: If True, assemble images (default: False)
    """
    with qt_app_context(exec_loop=True):
        items = make_items(N)
        _win = ptv.show_items(
            items,
            wintitle="Transform test ({}x{} images)".format(N, N),
            plot_type="image",
            show_itemlist=False,
        )
    if assemble_images:
        build_image(items)


if __name__ == "__main__":
    test_transform(N=500, assemble_images=True)



# content of conftest.py

import gc

import guidata
import h5py
import numpy
import PIL
import pytest
import qtpy
import qwt
import scipy
import tifffile
from guidata.env import execenv

import plotpy

# Turn on unattended mode for executing tests without user interaction
execenv.unattended = True
execenv.verbose = "quiet"


@pytest.fixture(scope="session", autouse=True)
def disable_gc_for_tests():
    """Disable garbage collection for all tests in the session."""
    # Important note:
    # ---------------
    # We need to disable garbage collection for all tests in the session because
    # this test suite is not representative of a typical application.
    # The high level of stress on the garbage collector can lead to false positives
    # in tests that rely on reference counting or finalization.
    # In a typical application, the garbage collector should be left enabled.

    gc.disable()
    yield
    gc.enable()


def pytest_report_header(config):
    """Add additional information to the pytest report header."""
    qtbindings_version = qtpy.PYSIDE_VERSION
    if qtbindings_version is None:
        qtbindings_version = qtpy.PYQT_VERSION
    return [
        f"PlotPy {plotpy.__version__}, guidata {guidata.__version__}, "
        f"PythonQwt {qwt.__version__}, "
        f"{qtpy.API_NAME} {qtbindings_version} [Qt version: {qtpy.QT_VERSION}]",
        f"NumPy {numpy.__version__}, SciPy {scipy.__version__}, "
        f"h5py {h5py.__version__}, "
        f"Pillow {PIL.__version__}, tifffile {tifffile.__version__}",
    ]



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""Selection tools"""

from __future__ import annotations

import numpy as np
from qtpy import QtCore as QC
from qtpy import QtGui as QG

from plotpy.config import _
from plotpy.events import (
    KeyEventMatch,
    ObjectHandler,
    StandardKeyMatch,
    StatefulEventFilter,
    setup_standard_tool_filter,
)
from plotpy.items import TrImageItem
from plotpy.plot import BasePlot
from plotpy.tools.base import InteractiveTool


class SelectTool(InteractiveTool):
    """
    Graphical Object Selection Tool
    """

    TITLE = _("Selection")
    ICON = "selection.png"
    CURSOR = QC.Qt.CursorShape.ArrowCursor

    def setup_filter(self, baseplot: BasePlot) -> int:
        """
        Set up the event filter for the tool.

        Args:
            baseplot: Base plot instance

        Returns:
            Start state of the filter
        """
        filter = baseplot.filter
        # Initialisation du filtre
        start_state = filter.new_state()
        # Bouton gauche :
        ObjectHandler(filter, QC.Qt.MouseButton.LeftButton, start_state=start_state)
        ObjectHandler(
            filter,
            QC.Qt.MouseButton.LeftButton,
            mods=QC.Qt.KeyboardModifier.ControlModifier,
            start_state=start_state,
            multiselection=True,
        )
        filter.add_event(
            start_state,
            KeyEventMatch(
                (QC.Qt.Key.Key_Enter, QC.Qt.Key.Key_Return, QC.Qt.Key.Key_Space)
            ),
            self.validate,
            start_state,
        )
        filter.add_event(
            start_state,
            StandardKeyMatch(QG.QKeySequence.SelectAll),
            self.select_all_items,
            start_state,
        )
        filter.add_event(
            start_state,
            KeyEventMatch(
                (
                    QC.Qt.Key.Key_Left,
                    QC.Qt.Key.Key_Right,
                    QC.Qt.Key.Key_Up,
                    QC.Qt.Key.Key_Down,
                )
            ),
            self.move_with_arrow,
            start_state,
        )
        filter.add_event(
            start_state,
            KeyEventMatch(
                [
                    (QC.Qt.Key.Key_Left, QC.Qt.KeyboardModifier.ControlModifier),
                    (QC.Qt.Key.Key_Right, QC.Qt.KeyboardModifier.ControlModifier),
                    (QC.Qt.Key.Key_Up, QC.Qt.KeyboardModifier.ControlModifier),
                    (QC.Qt.Key.Key_Down, QC.Qt.KeyboardModifier.ControlModifier),
                ]
            ),
            self.move_with_arrow,
            start_state,
        )
        filter.add_event(
            start_state,
            KeyEventMatch(
                [
                    (QC.Qt.Key.Key_Left, QC.Qt.KeyboardModifier.ShiftModifier),
                    (QC.Qt.Key.Key_Right, QC.Qt.KeyboardModifier.ShiftModifier),
                    (
                        QC.Qt.Key.Key_Left,
                        QC.Qt.KeyboardModifier.ControlModifier
                        | QC.Qt.KeyboardModifier.ShiftModifier,
                    ),
                    (
                        QC.Qt.Key.Key_Right,
                        QC.Qt.KeyboardModifier.ControlModifier
                        | QC.Qt.KeyboardModifier.ShiftModifier,
                    ),
                ]
            ),
            self.rotate_with_arrow,
            start_state,
        )
        return setup_standard_tool_filter(filter, start_state)

    def select_all_items(self, filter: StatefulEventFilter, event: QC.QEvent) -> None:
        """
        Select all items in the plot.

        Args:
            filter: Event filter
            event: Event that triggered the selection
        """
        filter.plot.select_all()
        filter.plot.replot()

    def move_with_arrow(self, filter: StatefulEventFilter, event: QG.QKeyEvent) -> None:
        """
        Move selected items with arrow keys.

        Args:
            filter: Event filter
            event: Key event
        """
        dx, dy = 0, 0
        if event.modifiers() == QC.Qt.KeyboardModifier.NoModifier:
            if event.key() == QC.Qt.Key.Key_Left:
                dx = -10
            elif event.key() == QC.Qt.Key.Key_Right:
                dx = 10
            elif event.key() == QC.Qt.Key.Key_Up:
                dy = -10
            elif event.key() == QC.Qt.Key.Key_Down:
                dy = 10
        elif event.modifiers() == QC.Qt.KeyboardModifier.ControlModifier:
            if event.key() == QC.Qt.Key.Key_Left:
                dx = -1
            elif event.key() == QC.Qt.Key.Key_Right:
                dx = 1
            elif event.key() == QC.Qt.Key.Key_Up:
                dy = -1
            elif event.key() == QC.Qt.Key.Key_Down:
                dy = 1
        selected_items = filter.plot.get_selected_items()
        for item in selected_items:
            if isinstance(item, TrImageItem):
                item.move_with_arrows(dx, dy)
        filter.plot.replot()

    def rotate_with_arrow(
        self, filter: StatefulEventFilter, event: QG.QKeyEvent
    ) -> None:
        """
        Rotate selected items with arrow keys.

        Args:
            filter: Event filter
            event: Key event
        """
        dangle = 0
        if event.modifiers() == QC.Qt.KeyboardModifier.ShiftModifier:
            if event.key() == QC.Qt.Key.Key_Left:
                dangle = -np.deg2rad(0.5)
            elif event.key() == QC.Qt.Key.Key_Right:
                dangle = np.deg2rad(0.5)
        elif (
            event.modifiers() & QC.Qt.KeyboardModifier.ControlModifier
            == QC.Qt.KeyboardModifier.ControlModifier
        ) and (
            event.modifiers() & QC.Qt.KeyboardModifier.ShiftModifier
            == QC.Qt.KeyboardModifier.ShiftModifier
        ):
            if event.key() == QC.Qt.Key.Key_Left:
                dangle = -np.deg2rad(0.05)
            elif event.key() == QC.Qt.Key.Key_Right:
                dangle = np.deg2rad(0.05)
        selected_items = filter.plot.get_selected_items()
        for item in selected_items:
            if isinstance(item, TrImageItem):
                item.rotate_with_arrows(dangle)
        filter.plot.replot()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
ColorMapEditor test

This plotpy widget can be used to edit and visualize a colormap. This test is used
to verify the consistancy of the colormap editor widget and some of the
CustomQwtLinearColormap features like initialization from a list of tuples or
modifications.
"""

# guitest: show

import qtpy.QtCore as QC
import qtpy.QtGui as QG
from guidata.qthelpers import qt_app_context

from plotpy.widgets.colormap.editor import ColorMapEditor
from plotpy.widgets.colormap.widget import EditableColormap


def test_colormap_editor() -> None:
    """Test the colormap editor widget and the CustomQwtLinearColormap class
    by using multiple methods to initialize and export the colormap.
    """
    with qt_app_context(exec_loop=True):
        print("Initialization of a default colormap editor widget")
        editor = ColorMapEditor(None)
        red = QG.QColor(QC.Qt.GlobalColor.red)
        green = QG.QColor(QC.Qt.GlobalColor.green)
        editor.colormap_widget.add_handle_at_relative_pos(0.5, red)
        editor.show()

        cmap_tuples = editor.get_colormap().to_tuples()
        print(
            "Initialization of a new colormap editor with the previous colormap: ",
            cmap_tuples,
        )
        new_cmap = EditableColormap.from_iterable(cmap_tuples)
        print(f"{new_cmap.to_tuples()}")
        editor = ColorMapEditor(None, colormap=new_cmap)
        editor.show()

        cmap_tuples = editor.get_colormap().to_tuples()
        print(
            "Initialization of a new default colormap editor, "
            "modified post-initialization with the previous colormap: ",
            cmap_tuples,
        )
        new_cmap = EditableColormap.from_iterable(cmap_tuples)
        editor = ColorMapEditor(None)
        editor.set_colormap(new_cmap)
        editor.show()

        cmap_tuples = editor.get_colormap().to_tuples()
        cmap_tuples = tuple((int(val * 255 + 1), color) for val, color in cmap_tuples)
        print(
            "Initialization of a new default colormap editor, "
            "modified post-initialization with the previous colormap "
            "with stops scaled by 255 + 1: ",
            cmap_tuples,
        )
        new_cmap = EditableColormap.from_iterable(cmap_tuples)
        editor = ColorMapEditor(None)
        editor.set_colormap(new_cmap)
        editor.show()

        print(
            "Initialization of a new default colormap editor, "
            "modified post-initialization with the previous colormap "
            "where the red stop is replaced with a green stop: ",
            cmap_tuples,
        )

        editor = ColorMapEditor(None)
        editor.set_colormap(new_cmap)
        editor.colormap_widget.edit_color_stop(1, None, green)
        editor.show()


if __name__ == "__main__":
    test_colormap_editor()



# -*- coding: utf-8 -*-

from __future__ import annotations

import math
from typing import TYPE_CHECKING

import numpy as np
from guidata.configtools import get_icon
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG

from plotpy.config import CONF, _
from plotpy.items.shape.polygon import PolygonShape
from plotpy.styles.shape import AxesShapeParam

if TYPE_CHECKING:
    import guidata.io
    import qwt.scale_map
    from qtpy.QtCore import QRectF
    from qtpy.QtGui import QPainter

    from plotpy.styles.base import ItemParameters
    from plotpy.styles.shape import ShapeParam


class Axes(PolygonShape):
    """Axes shape

    Args:
        p0: First point (0,1)
        p1: Second point (1,1)
        p2: Third point (0,0)
        axesparam: Axes parameters
        shapeparam: Shape parameters
    """

    CLOSED: bool = True

    def __init__(
        self,
        p0: tuple[float, float] = (0, 0),
        p1: tuple[float, float] = (0, 0),
        p2: tuple[float, float] = (0, 0),
        axesparam: AxesShapeParam = None,
        shapeparam: ShapeParam = None,
    ) -> None:
        super().__init__(shapeparam=shapeparam)
        self.set_rect(p0, p1, p2)
        self.arrow_angle = 15  # degrees
        self.arrow_size = 0.05  # % of axe length
        self.x_pen = self.pen
        self.x_brush = self.brush
        self.y_pen = self.pen
        self.y_brush = self.brush
        if axesparam is None:
            self.axesparam = AxesShapeParam(_("Axes"), icon="gtaxes.png")
        else:
            self.axesparam = axesparam
        self.axesparam.update_param(self)
        self.setIcon(get_icon("gtaxes.png"))

    def __reduce__(self) -> tuple:
        """Reduce object to picklable state"""
        self.axesparam.update_param(self)
        state = (self.shapeparam, self.axesparam, self.points, self.z())
        return (self.__class__, (), state)

    def __setstate__(self, state: tuple) -> None:
        """Set object state from pickled state"""
        shapeparam, axesparam, points, z = state
        self.points = points
        self.setZ(z)
        self.shapeparam: ShapeParam = shapeparam
        self.shapeparam.update_item(self)
        self.axesparam: AxesShapeParam = axesparam
        self.axesparam.update_item(self)

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        super().serialize(writer)
        self.axesparam.update_param(self)
        writer.write(self.axesparam, group_name="axesparam")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        super().deserialize(reader)
        self.axesparam = AxesShapeParam(_("Axes"), icon="gtaxes.png")
        reader.read("axesparam", instance=self.axesparam)
        self.axesparam.update_item(self)

    def get_transform_matrix(self, dx: float = 1.0, dy: float = 1.0) -> np.ndarray:
        """Return the transformation matrix

        Args:
            dx: X scale
            dy: Y scale

        Returns:
            Transformation matrix
        """
        p0, p1, _p3, p2 = [np.array([p[0], p[1], 1.0]) for p in self.points]
        matrix = np.array([(p1 - p0) / dx, (p2 - p0) / dy, p0])
        if abs(np.linalg.det(matrix)) > 1e-12:
            return np.linalg.inv(matrix)

    def set_rect(
        self, p0: tuple[float, float], p1: tuple[float, float], p2: tuple[float, float]
    ) -> None:
        """Set the coordinates of the axes

        Args:
            p0: First point (0,1)
            p1: Second point (1,1)
            p2: Third point (0,0)
        """
        p3x = p1[0] + p2[0] - p0[0]
        p3y = p1[1] + p2[1] - p0[1]
        self.set_points([p0, p1, (p3x, p3y), p2])

    def set_style(self, section: str, option: str) -> None:
        """Set style for this item

        Args:
            section: Section
            option: Option
        """
        PolygonShape.set_style(self, section, option + "/border")
        self.axesparam.read_config(CONF, section, option)
        self.axesparam.update_item(self)

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        _nx, _ny = pos
        p0, p1, _p3, p2 = list(self.points)
        d1x = p1[0] - p0[0]
        d1y = p1[1] - p0[1]
        d2x = p2[0] - p0[0]
        d2y = p2[1] - p0[1]
        if handle == 0:
            pp0 = pos
            pp1 = pos[0] + d1x, pos[1] + d1y
            pp2 = pos[0] + d2x, pos[1] + d2y
        elif handle == 1:
            pp0 = p0
            pp1 = pos
            pp2 = p2
        elif handle == 3:
            pp0 = p0
            pp1 = p1
            pp2 = pos
        elif handle == 2:
            # find (a,b) such that p3 = a*d1 + b*d2 + p0
            d3x = pos[0] - p0[0]
            d3y = pos[1] - p0[1]
            det = d1x * d2y - d2x * d1y
            if abs(det) < 1e-6:
                # reset
                d1x = d2y = 1.0
                d1y = d2x = 0.0
                det = 1.0
            a = (d2y * d3x - d2x * d3y) / det
            b = (-d1y * d3x + d1x * d3y) / det
            _pp3 = pos
            pp1 = p0[0] + a * d1x, p0[1] + a * d1y
            pp2 = p0[0] + b * d2x, p0[1] + b * d2y
            pp0 = p0
        self.set_rect(pp0, pp1, pp2)
        if self.plot():
            self.plot().SIG_AXES_CHANGED.emit(self)

    def move_shape(self, old_pos: QC.QPointF, new_pos: QC.QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in axis coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        PolygonShape.move_shape(self, old_pos, new_pos)
        if self.plot():
            self.plot().SIG_AXES_CHANGED.emit(self)

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        PolygonShape.draw(self, painter, xMap, yMap, canvasRect)
        p0, p1, _, p2 = list(self.points)

        points = self.transform_points(xMap, yMap)  # points is a QPolygonF

        painter.setPen(self.x_pen)
        painter.setBrush(self.x_brush)
        painter.drawLine(points.at(0), points.at(1))
        self.draw_arrow(painter, xMap, yMap, p0, p1)
        painter.setPen(self.y_pen)
        painter.setBrush(self.y_brush)
        painter.drawLine(points.at(0), points.at(3))
        self.draw_arrow(painter, xMap, yMap, p0, p2)

    def draw_arrow(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        p0: tuple[float, float],
        p1: tuple[float, float],
    ) -> None:
        """Draw an arrow

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            p0: First point
            p1: Second point
        """
        sz = self.arrow_size
        angle = math.pi * self.arrow_angle / 180.0
        ca, sa = math.cos(angle), math.sin(angle)
        d1x = xMap.transform(p1[0]) - xMap.transform(p0[0])
        d1y = yMap.transform(p1[1]) - yMap.transform(p0[1])
        norm = math.sqrt(d1x**2 + d1y**2)
        if abs(norm) < 1e-6:
            return
        d1x *= sz / norm
        d1y *= sz / norm
        n1x = -d1y
        n1y = d1x
        # arrow : a0 - a1 == p1 - a2
        a1x = xMap.transform(p1[0])
        a1y = yMap.transform(p1[1])
        a0x = a1x - ca * d1x + sa * n1x
        a0y = a1y - ca * d1y + sa * n1y
        a2x = a1x - ca * d1x - sa * n1x
        a2y = a1y - ca * d1y - sa * n1y

        poly = QG.QPolygonF()
        poly.append(QC.QPointF(a0x, a0y))
        poly.append(QC.QPointF(a1x, a1y))
        poly.append(QC.QPointF(a2x, a2y))
        painter.drawPolygon(poly)

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.axesparam.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        PolygonShape.get_item_parameters(self, itemparams)
        self.update_item_parameters()
        itemparams.add("AxesShapeParam", self, self.axesparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        PolygonShape.set_item_parameters(self, itemparams)
        update_dataset(
            self.axesparam, itemparams.get("AxesShapeParam"), visible_only=True
        )
        self.axesparam.update_item(self)


assert_interfaces_valid(Axes)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
plotpy.widgets.label
--------------------

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np
from guidata.configtools import get_icon
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qwt import QwtPlotItem

from plotpy.config import CONF, _
from plotpy.coords import canvas_to_axes
from plotpy.interfaces import IBasePlotItem, ISerializableType, IShapeItemType
from plotpy.items.curve.base import CurveItem
from plotpy.styles.label import LabelParam

if TYPE_CHECKING:
    from collections.abc import Callable

    import guidata.io
    import qwt.scale_map
    import qwt.symbol
    from qtpy.QtCore import QPointF, QRectF
    from qtpy.QtGui import QBrush, QPainter, QPen, QTextDocument

    from plotpy.interfaces import IItemType
    from plotpy.items import ImageItem, RectangleShape, XRangeSelection
    from plotpy.styles.base import ItemParameters

ANCHORS = {
    "TL": lambda r: (r.left(), r.top()),
    "TR": lambda r: (r.right(), r.top()),
    "BL": lambda r: (r.left(), r.bottom()),
    "BR": lambda r: (r.right(), r.bottom()),
    "L": lambda r: (r.left(), (r.top() + r.bottom()) / 2.0),
    "R": lambda r: (r.right(), (r.top() + r.bottom()) / 2.0),
    "T": lambda r: ((r.left() + r.right()) / 2.0, r.top()),
    "B": lambda r: ((r.left() + r.right()) / 2.0, r.bottom()),
    "C": lambda r: ((r.left() + r.right()) / 2.0, (r.top() + r.bottom()) / 2.0),
}


class AbstractLabelItem(QwtPlotItem):
    """Abstract label plot item

    Draws a label on the canvas at position :
    G+C where G is a point in plot coordinates and C a point
    in canvas coordinates.
    G can also be an anchor string as in ANCHORS in which case
    the label will keep a fixed position wrt the canvas rect

    Args:
        labelparam: Label parameters
    """

    _readonly = False
    _private = False

    def __init__(self, labelparam: LabelParam = None) -> None:
        super().__init__()
        self.selected = False
        self.anchor = None
        self.G = None
        self.C = None
        self.border_pen = None
        self.bg_brush = None
        if labelparam is None:
            self.labelparam = LabelParam(_("Label"), icon="label.png")
        else:
            self.labelparam = labelparam
            self.labelparam.update_item(self)
        self._can_select = True
        self._can_resize = False
        self._can_move = True
        self._can_rotate = False

    def set_style(self, section: str, option: str) -> None:
        """Set style for this item

        Args:
            section: Section
            option: Option
        """
        self.labelparam.read_config(CONF, section, option)
        self.labelparam.update_item(self)

    def __reduce__(self) -> tuple[type, tuple]:
        """Return a tuple containing the constructor and its arguments"""
        return (self.__class__, (self.labelparam,))

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        self.labelparam.update_param(self)
        writer.write(self.labelparam, group_name="labelparam")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        self.labelparam = LabelParam(_("Label"), icon="label.png")
        reader.read("labelparam", instance=self.labelparam)
        self.labelparam.update_item(self)
        if isinstance(self.G, np.ndarray):
            self.G = tuple(self.G)

    def get_text_rect(self) -> QC.QRectF:
        """Return the text rectangle

        Returns:
            Text rectangle
        """
        return QC.QRectF(0.0, 0.0, 10.0, 10.0)

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IShapeItemType,)

    def set_text_style(
        self, font: QG.QFont | None = None, color: str | None = None
    ) -> None:
        """Set label text style

        Args:
            font: Font
            color: Color
        """
        raise NotImplementedError

    def get_top_left(
        self,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QC.QRectF,
    ) -> tuple[float, float]:
        """Return the top left corner of the text rectangle

        Args:
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle

        Returns:
            tuple: Tuple with two elements: (x, y)
        """
        x0, y0 = self.get_origin(xMap, yMap, canvasRect)
        x0 += self.C[0]
        y0 += self.C[1]

        rect = self.get_text_rect()
        pos = ANCHORS[self.anchor](rect)
        x0 -= pos[0]
        y0 -= pos[1]
        return x0, y0

    def get_origin(
        self,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QC.QRectF,
    ) -> tuple[float, float]:
        """Return the origin of the text rectangle

        Args:
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle

        Returns:
            tuple: Tuple with two elements: (x, y)
        """
        if self.G in ANCHORS:
            return ANCHORS[self.G](canvasRect)
        else:
            x0 = xMap.transform(self.G[0])
            y0 = yMap.transform(self.G[1])
            return x0, y0

    def set_selectable(self, state: bool) -> None:
        """Set item selectable state

        Args:
            state: True if item is selectable, False otherwise
        """
        self._can_select = state

    def set_resizable(self, state: bool) -> None:
        """Set item resizable state
        (or any action triggered when moving an handle, e.g. rotation)

        Args:
            state: True if item is resizable, False otherwise
        """
        self._can_resize = state

    def set_movable(self, state: bool) -> None:
        """Set item movable state

        Args:
            state: True if item is movable, False otherwise
        """
        self._can_move = state

    def set_rotatable(self, state: bool) -> None:
        """Set item rotatable state

        Args:
            state: True if item is rotatable, False otherwise
        """
        self._can_rotate = state

    def can_select(self) -> bool:
        """
        Returns True if this item can be selected

        Returns:
            bool: True if item can be selected, False otherwise
        """
        return self._can_select

    def can_resize(self) -> bool:
        """
        Returns True if this item can be resized

        Returns:
            bool: True if item can be resized, False otherwise
        """
        return self._can_resize

    def can_move(self) -> bool:
        """
        Returns True if this item can be moved

        Returns:
            bool: True if item can be moved, False otherwise
        """
        return self._can_move

    def can_rotate(self) -> bool:
        """
        Returns True if this item can be rotated

        Returns:
            bool: True if item can be rotated, False otherwise
        """
        return False  # TODO: Implement labels rotation?

    def set_readonly(self, state: bool) -> None:
        """Set object readonly state

        Args:
            state: True if object is readonly, False otherwise
        """
        self._readonly = state

    def is_readonly(self) -> bool:
        """Return object readonly state

        Returns:
            bool: True if object is readonly, False otherwise
        """
        return self._readonly

    def set_private(self, state: bool) -> None:
        """Set object as private

        Args:
            state: True if object is private, False otherwise
        """
        self._private = state

    def is_private(self) -> bool:
        """Return True if object is private

        Returns:
            bool: True if object is private, False otherwise
        """
        return self._private

    def invalidate_plot(self) -> None:
        """Invalidate the plot to force a redraw"""
        plot = self.plot()
        if plot:
            plot.invalidate()

    def select(self) -> None:
        """
        Select the object and eventually change its appearance to highlight the
        fact that it's selected
        """
        if self.selected:
            # Already selected
            return
        self.selected = True
        w = self.border_pen.width()
        self.border_pen.setWidth(w + 1)
        self.invalidate_plot()

    def unselect(self) -> None:
        """
        Unselect the object and eventually restore its original appearance to
        highlight the fact that it's not selected anymore
        """
        self.selected = False
        self.labelparam.update_item(self)
        self.invalidate_plot()

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        plot = self.plot()
        if plot is None:
            return
        rect = self.get_text_rect()
        canvasRect = plot.canvas().contentsRect()
        xMap = plot.canvasMap(self.xAxis())
        yMap = plot.canvasMap(self.yAxis())
        x, y = self.get_top_left(xMap, yMap, canvasRect)
        rct = QC.QRectF(x, y, rect.width(), rect.height())
        inside = rct.contains(pos.x(), pos.y())
        if inside:
            return self.click_inside(pos.x() - x, pos.y() - y)
        else:
            return 1000.0, None, False, None

    def click_inside(self, locx: float, locy: float) -> tuple[float, float, bool, type]:
        """Called when the mouse button is clicked inside the object

        Args:
            locx: Local x coordinate
            locy: Local y coordinate

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).
        """
        return 2.0, 1, True, None

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.labelparam.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        self.update_item_parameters()
        itemparams.add("LabelParam", self, self.labelparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(self.labelparam, itemparams.get("LabelParam"), visible_only=True)
        self.labelparam.update_item(self)
        if self.selected:
            self.select()

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        if handle != -1:
            return

    def move_local_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in canvas coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        plot = self.plot()
        if plot is None:
            return
        if self.G in ANCHORS or not self.labelparam.move_anchor:
            # Move canvas offset
            lx, ly = self.C
            lx += new_pos.x() - old_pos.x()
            ly += new_pos.y() - old_pos.y()
            self.C = lx, ly
            self.labelparam.xc, self.labelparam.yc = lx, ly
            lx0, ly0 = canvas_to_axes(self, old_pos)
            lx1, ly1 = canvas_to_axes(self, new_pos)
        else:
            # Move anchor
            lx0, ly0 = self.G
            cx = plot.transform(self.xAxis(), lx0)
            cy = plot.transform(self.yAxis(), ly0)
            cx += new_pos.x() - old_pos.x()
            cy += new_pos.y() - old_pos.y()
            lx1 = plot.invTransform(self.xAxis(), cx)
            ly1 = plot.invTransform(self.yAxis(), cy)
            self.G = lx1, ly1
            self.labelparam.xg, self.labelparam.yg = lx1, ly1
        plot.SIG_ITEM_MOVED.emit(self, lx0, ly0, lx1, ly1)

    def move_with_selection(self, delta_x: float, delta_y: float) -> None:
        """Translate the item together with other selected items

        Args:
            delta_x: Translation in plot coordinates along x-axis
            delta_y: Translation in plot coordinates along y-axis
        """
        if self.G in ANCHORS or not self.labelparam.move_anchor:
            return
        lx0, ly0 = self.G
        lx1, ly1 = lx0 + delta_x, ly0 + delta_y
        self.G = lx1, ly1
        self.labelparam.xg, self.labelparam.yg = lx1, ly1

    def draw_frame(
        self, painter: QPainter, x: float, y: float, w: float, h: float
    ) -> None:
        """Draw the frame

        Args:
            painter: Painter
            x: X coordinate
            y: Y coordinate
            w: Width
            h: Height
        """
        rectf = QC.QRectF(x, y, w, h)
        if self.labelparam.bgalpha > 0.0:
            painter.fillRect(rectf, self.bg_brush)
        if self.border_pen.width() > 0:
            painter.setPen(self.border_pen)
            painter.drawRect(rectf)


class LabelItem(AbstractLabelItem):
    """Label plot item

    Args:
        text: Text
        labelparam: Label parameters
    """

    __implements__ = (IBasePlotItem, ISerializableType)

    def __init__(
        self, text: str | None = None, labelparam: LabelParam | None = None
    ) -> None:
        self.text_string = "" if text is None else text
        self.text = QG.QTextDocument()
        self.marker: qwt.symbol.QwtSymbol | None = None
        super().__init__(labelparam)
        self.setIcon(get_icon("label.png"))

    def __reduce__(self) -> tuple[type, tuple]:
        """Return a tuple containing the constructor and its arguments"""
        return (self.__class__, (self.text_string, self.labelparam))

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        super().serialize(writer)
        writer.write(self.text_string, group_name="text")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        super().deserialize(reader)
        self.set_text(reader.read("text", func=reader.read_unicode))

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IShapeItemType, ISerializableType)

    def set_pos(self, x: float, y: float) -> None:
        """Set position

        Args:
            x: X coordinate
            y: Y coordinate
        """
        self.G = x, y
        self.labelparam.abspos = False
        self.labelparam.xg, self.labelparam.yg = x, y

    def get_plain_text(self) -> str:
        """Get plain text

        Returns:
            str: Plain text
        """
        return str(self.text.toPlainText())

    def set_text(self, text: str | None = None) -> None:
        """Set text

        Args:
            text: Text
        """
        if text is not None:
            self.text_string = text
        self.text.setHtml(f"<div>{self.text_string}</div>")

    def set_text_style(
        self, font: QG.QFont | None = None, color: str | None = None
    ) -> None:
        """Set label text style

        Args:
            font: Font
            color: Color
        """
        if font is not None:
            self.text.setDefaultFont(font)
        if color is not None:
            self.text.setDefaultStyleSheet("div { color: %s; }" % color)
        self.set_text()

    def get_text_rect(self) -> QRectF:
        """Return the text rectangle

        Returns:
            Text rectangle
        """
        sz = self.text.size()
        return QC.QRectF(0, 0, sz.width(), sz.height())

    def update_text(self) -> None:
        """Update text"""
        pass

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        self.update_text()
        x, y = self.get_top_left(xMap, yMap, canvasRect)
        x0, y0 = self.get_origin(xMap, yMap, canvasRect)
        painter.save()
        self.marker.drawSymbols(painter, [QC.QPointF(x0, y0)])
        painter.restore()
        sz = self.text.size()
        self.draw_frame(painter, int(x), int(y), int(sz.width()), int(sz.height()))
        painter.setPen(QG.QPen(QG.QColor(self.labelparam.color)))
        painter.translate(x, y)
        self.text.drawContents(painter)


assert_interfaces_valid(LabelItem)


LEGEND_WIDTH = 30  # Length of the sample line
LEGEND_SPACEH = 5  # Spacing between border, sample, text, border
LEGEND_SPACEV = 3  # Vertical space between items


class LegendBoxItem(AbstractLabelItem):
    """Legend box plot item

    Args:
        labelparam: Label parameters
    """

    __implements__ = (IBasePlotItem, ISerializableType)

    def __init__(self, labelparam: LabelParam = None) -> None:
        self.font = None
        self.color = None
        super().__init__(labelparam)
        # saves the last computed sizes
        self.sizes = 0.0, 0.0, 0.0, 0.0
        self.setIcon(get_icon("legend.png"))

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IShapeItemType, ISerializableType)

    def get_legend_items(
        self,
    ) -> list[tuple[QTextDocument, QPen, QBrush, qwt.symbol.QwtSymbol]]:
        """Return the legend items

        Returns:
            list: List of legend items (text, pen, brush, symbol)
        """
        plot = self.plot()
        if plot is None:
            return []
        text_items = []
        for item in plot.get_items():
            if not isinstance(item, CurveItem) or not self.include_item(item):
                continue
            text = QG.QTextDocument()
            text.setDefaultFont(self.font)
            text.setDefaultStyleSheet("div { color: %s; }" % self.color)
            text.setHtml(f"<div>{item.param.label}</div>")
            text_items.append((text, item.pen(), item.brush(), item.symbol()))
        return text_items

    def include_item(self, item: Any) -> bool:
        """Include item in legend box?

        Args:
            item: Item

        Returns:
            True if item is included, False otherwise
        """
        return item.isVisible()

    def get_legend_size(
        self, legenditems: list[tuple]
    ) -> tuple[float, float, float, float]:
        """Return the legend size

        Args:
            legenditems: Legend items

        Returns:
            tuple: Tuple with four elements: (TW, TH, width, height)
        """
        width = 0
        height = 0
        for text, _, _, _ in legenditems:  # noqa
            sz = text.size()
            if sz.width() > width:
                width = sz.width()
            if sz.height() > height:
                height = sz.height()

        TW = LEGEND_SPACEH * 3 + LEGEND_WIDTH + width
        TH = len(legenditems) * (height + LEGEND_SPACEV) + LEGEND_SPACEV
        self.sizes = TW, TH, width, height
        return self.sizes

    def set_text_style(
        self, font: QG.QFont | None = None, color: str | None = None
    ) -> None:
        """Set label text style

        Args:
            font: Font
            color: Color
        """
        if font is not None:
            self.font = font
        if color is not None:
            self.color = color

    def get_text_rect(self) -> QC.QRectF:
        """Return the text rectangle

        Returns:
            Text rectangle
        """
        items = self.get_legend_items()
        TW, TH, _width, _height = self.get_legend_size(items)
        return QC.QRectF(0.0, 0.0, TW, TH)

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        items = self.get_legend_items()
        TW, TH, _width, height = self.get_legend_size(items)

        x, y = self.get_top_left(xMap, yMap, canvasRect)
        self.draw_frame(painter, int(x), int(y), int(TW), int(TH))

        y0 = int(y + LEGEND_SPACEV)
        x0 = int(x + LEGEND_SPACEH)
        for text, ipen, ibrush, isymbol in items:
            isymbol.drawSymbols(
                painter, [QC.QPointF(x0 + LEGEND_WIDTH / 2, y0 + height / 2)]
            )
            painter.save()
            painter.setPen(ipen)
            painter.setBrush(ibrush)
            painter.drawLine(
                x0, int(y0 + height / 2), int(x0 + LEGEND_WIDTH), int(y0 + height / 2)
            )
            x1 = x0 + LEGEND_SPACEH + LEGEND_WIDTH
            painter.translate(x1, y0)
            text.drawContents(painter)
            painter.restore()
            y0 += height + LEGEND_SPACEV

    def click_inside(self, locx: float, locy: float) -> tuple[float, float, bool, type]:
        """Called when the mouse button is clicked inside the object

        Args:
            locx: Local x coordinate
            locy: Local y coordinate

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).
        """
        # hit_test already called get_text_rect for us...
        _TW, _TH, _width, height = self.sizes
        line = (locy - LEGEND_SPACEV) / (height + LEGEND_SPACEV)
        line = int(line)
        if LEGEND_SPACEH <= locx <= (LEGEND_WIDTH + LEGEND_SPACEH):
            # We hit a legend line, select the corresponding curve
            # and do as if we weren't hit...
            items = [
                item
                for item in self.plot().get_items()
                if self.include_item(item) and isinstance(item, CurveItem)
            ]
            if line < len(items):
                return 1000.0, None, False, items[line]
        return 2.0, 1, True, None

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.labelparam.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        self.update_item_parameters()
        itemparams.add("LegendParam", self, self.labelparam)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(
            self.labelparam, itemparams.get("LegendParam"), visible_only=True
        )
        self.labelparam.update_item(self)
        if self.selected:
            self.select()


assert_interfaces_valid(LegendBoxItem)


class SelectedLegendBoxItem(LegendBoxItem):
    """Selected legend box plot item

    Args:
        labelparam: Label parameters
        itemlist: List of items
    """

    def __init__(
        self, dataset: LabelParam = None, itemlist: list[Any] | None = None
    ) -> None:
        super().__init__(dataset)
        self.itemlist = [] if itemlist is None else itemlist

    def __reduce__(self) -> tuple[type, tuple]:
        """Return a tuple containing the constructor and its arguments"""
        # XXX Filter itemlist for picklabel items
        return (self.__class__, (self.labelparam, []))

    def include_item(self, item: Any) -> bool:
        """Include item in legend box?

        Args:
            item: Item

        Returns:
            True if item is included, False otherwise
        """
        return LegendBoxItem.include_item(self, item) and item in self.itemlist

    def add_item(self, item: Any) -> None:
        """Add item

        Args:
            item: Item
        """
        self.itemlist.append(item)


class ObjectInfo:
    """Base class for objects used to format text labels"""

    def get_text(self) -> str:
        """Return the text to be displayed"""
        return ""


class RangeInfo(ObjectInfo):
    """ObjectInfo handling XRangeSelection shape informations: x, dx

    Args:
        label: Formatted string
        xrangeselection: XRangeSelection object
        function: Input arguments are x, dx ; returns objects used to format the
         label. Default function is `lambda x, dx: (x, dx)`.

    Examples::
        x = linspace(-10, 10, 10)
        y = sin(sin(sin(x)))
        xrangeselection = make.range(-2, 2)
        RangeInfo(u"x = %.1f ± %.1f cm", xrangeselection, lambda x, dx: (x, dx))
        disp = make.info_label('BL', comp, title="titre")
    """

    def __init__(
        self,
        label: str,
        xrangeselection: XRangeSelection,
        function: Callable | None = None,
    ) -> None:
        self.label = str(label)
        self.range = xrangeselection
        if function is None:

            def function(x, dx):
                return x, dx

        self.func = function

    def get_text(self) -> str:
        """Return the text to be displayed"""
        x0, x1 = self.range.get_range()
        x = 0.5 * (x0 + x1)
        dx = 0.5 * (x1 - x0)
        return self.label % self.func(x, dx)


class RangeComputation(ObjectInfo):
    """ObjectInfo showing curve computations relative to a XRangeSelection
    shape

    Args:
        label: formatted string
        curve: CurveItem object
        xrangeselection: XRangeSelection object
        function: input arguments are x, y arrays (extraction of arrays
         corresponding to the xrangeselection X-axis range)
    """

    def __init__(
        self,
        label: str,
        curve: CurveItem,
        xrangeselection: XRangeSelection,
        function: Callable | None = None,
    ) -> None:
        self.label = str(label)
        self.curve = curve
        self.range = xrangeselection
        if function is None:

            def function(x, dx):
                return x, dx

        self.func = function

    def set_curve(self, curve: CurveItem) -> None:
        """Set curve item

        Args:
            curve: Curve item
        """
        self.curve = curve

    def get_text(self) -> str:
        """Return the text to be displayed"""
        x0, x1 = self.range.get_range()
        data = self.curve.get_data()
        X = data[0]
        i0 = X.searchsorted(x0)
        i1 = X.searchsorted(x1)
        if i0 > i1:
            i0, i1 = i1, i0
        vectors = []
        for vector in data:
            if vector is None:
                vectors.append(None)
            elif i0 == i1:
                vectors.append(np.array([np.nan]))
            else:
                vectors.append(vector[i0:i1])
        return self.label % self.func(*vectors)


class RangeComputation2d(ObjectInfo):
    """ObjectInfo showing image computations relative to a rectangular area

    Args:
        label: formatted string
        image: ImageItem object
        rect: Rectangular area
        function: input arguments are x, y, z arrays (extraction of arrays
         corresponding to the rectangular area)
    """

    def __init__(
        self, label: str, image: ImageItem, rect: RectangleShape, function: Callable
    ) -> None:
        self.label = str(label)
        self.image = image
        self.rect = rect
        self.func = function

    def get_text(self) -> str:
        """Return the text to be displayed"""
        x0, y0, x1, y1 = self.rect.get_rect()
        x, y, z = self.image.get_data(x0, y0, x1, y1)
        res = self.func(x, y, z)
        return self.label % res


class DataInfoLabel(LabelItem):
    """Label item displaying informations relative to an annotation

    Args:
        labelparam: Label parameters
        infos: List of objects implementing the ``get_text`` method
    """

    __implements__ = (IBasePlotItem,)

    def __init__(
        self, labelparam: LabelParam | None = None, infos: list[Any] | None = None
    ) -> None:
        super().__init__(None, labelparam)
        if isinstance(infos, ObjectInfo):
            infos = [infos]
        self.infos = infos

    def __reduce__(self) -> tuple[type, tuple]:
        """Return a tuple containing the constructor and its arguments"""
        return (self.__class__, (self.labelparam, self.infos))

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IShapeItemType,)

    def update_text(self):
        """Update text"""
        title = self.labelparam.label
        if title:
            text = ["<b>%s</b>" % title]
        else:
            text = []
        for info in self.infos:
            text.append(info.get_text())
        self.set_text("<br/>".join(text))



import numpy as np
import pytest
import qtpy.QtCore as QC

from plotpy.items.curve.base import seg_dist, seg_dist_v

param_list = [
    ((200, 100), (150, 196), (250, 180), 86),
    ((200, 100), (190, 196), (210, 180), 80),
    ((201, 105), (201, 196), (201, 180), 75),
]


@pytest.mark.parametrize("point_1,point_2,point_3,output", param_list)
def test_seg_dist(point_1, point_2, point_3, output):
    """ """
    ret = seg_dist(QC.QPointF(*point_1), QC.QPointF(*point_2), QC.QPointF(*point_3))
    assert int(ret) == output


def test_seg_dist_v():
    """Test de seg_dist_v"""
    a = (np.arange(10.0) ** 2).reshape(5, 2)
    ix, dist = seg_dist_v((2.1, 3.3), a[:-1, 0], a[:-1, 1], a[1:, 0], a[1:, 1])
    assert ix == 0
    assert round(dist, 2) == 0.85


if __name__ == "__main__":
    test_seg_dist_v()



# -*- coding: utf-8 -*-

"""
Colormaps
=========

The `colormap` module contains definition of common colormaps and tools
to manipulate and create them.

.. automodule:: plotpy.widgets.colormap.widget
    :members:
.. automodule:: plotpy.widgets.colormap.editor
    :members:
.. automodule:: plotpy.widgets.colormap.manager
    :members:
"""



# -*- coding: utf-8 -*-

import os
import os.path as osp
import platform
import sys

import numpy
from Cython.Compiler import Main
from setuptools import Distribution, Extension, setup

LIBNAME = "plotpy"
SRCPATH = osp.join(".", "src")


def is_msvc():
    """Detect if Microsoft Visual C++ compiler was chosen to build package"""
    dist = Distribution()
    dist.parse_config_files()
    bld = dist.get_option_dict("build")
    if bld:
        comp = bld.get("compiler")
        if comp is not None and "mingw32" in comp:
            return False  # mingw is the compiler
    return os.name == "nt" and "mingw" not in "".join(sys.argv)


def compile_cython_extensions():
    """Compile Cython extensions"""
    for fname in os.listdir(SRCPATH):
        if osp.splitext(fname)[1] == ".pyx":
            Main.compile(osp.join(SRCPATH, fname), language_level=2)


compile_cython_extensions()

INCLUDE_DIRS = [SRCPATH, numpy.get_include()]

MACROS_CYTHON = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
CFLAGS_CYTHON = []
MACROS_CPP = [("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
CFLAGS_CPP = ["/EHsc", "/fp:fast"] if is_msvc() else ["-Wall"]
if platform.system() == "Darwin":
    CFLAGS_CPP += ["-std=c++11"]

setup(
    ext_modules=[
        Extension(
            name=f"{LIBNAME}.mandelbrot",
            sources=[osp.join(SRCPATH, "mandelbrot.c")],
            include_dirs=INCLUDE_DIRS,
            extra_compile_args=CFLAGS_CYTHON,
            define_macros=MACROS_CYTHON,
        ),
        Extension(
            name=f"{LIBNAME}.histogram2d",
            sources=[osp.join(SRCPATH, "histogram2d.c")],
            include_dirs=INCLUDE_DIRS,
            extra_compile_args=CFLAGS_CYTHON,
            define_macros=MACROS_CYTHON,
        ),
        Extension(
            name=f"{LIBNAME}._scaler",
            sources=[osp.join(SRCPATH, "scaler.cpp"), osp.join(SRCPATH, "pcolor.cpp")],
            extra_compile_args=CFLAGS_CPP,
            depends=[
                osp.join(SRCPATH, "traits.hpp"),
                osp.join(SRCPATH, "points.hpp"),
                osp.join(SRCPATH, "arrays.hpp"),
                osp.join(SRCPATH, "scaler.hpp"),
                osp.join(SRCPATH, "debug.hpp"),
            ],
            include_dirs=INCLUDE_DIRS,
            define_macros=MACROS_CPP,
        ),
    ]
)



# -*- coding: utf-8 -*-

# pylint: disable=W0611
# flake8: noqa

from plotpy.styles.axes import (
    AxesParam,
    AxeStyleParam,
    AxisItem,
    AxisItemWidget,
    AxisParam,
    ImageAxesParam,
)
from plotpy.styles.base import (
    COLORS,
    LINESTYLES,
    MARKERS,
    BrushStyleItem,
    BrushStyleItemWidget,
    BrushStyleParam,
    FontItem,
    FontItemWidget,
    FontParam,
    GridParam,
    ItemParameters,
    LineStyleItem,
    LineStyleItemWidget,
    LineStyleParam,
    SymbolItem,
    SymbolItemWidget,
    SymbolParam,
    TextStyleItem,
    TextStyleItemWidget,
    TextStyleParam,
    style_generator,
    update_style_attr,
)
from plotpy.styles.curve import CurveParam, CurveParam_MS
from plotpy.styles.errorbar import ErrorBarParam
from plotpy.styles.histogram import (
    Histogram2DParam,
    Histogram2DParam_MS,
    HistogramParam,
)
from plotpy.styles.image import (
    BaseImageParam,
    ImageFilterParam,
    ImageParam,
    ImageParam_MS,
    MaskedImageParam,
    MaskedImageParam_MS,
    MaskedImageParamMixin,
    MaskedXYImageParam,
    MaskedXYImageParam_MS,
    QuadGridParam,
    RawImageParam,
    RawImageParam_MS,
    RGBImageParam,
    TrImageParam,
    TrImageParam_MS,
    XYImageParam,
    XYImageParam_MS,
)
from plotpy.styles.label import (
    LabelParam,
    LabelParam_MS,
    LabelParamWithContents,
    LabelParamWithContents_MS,
    LegendParam,
    LegendParam_MS,
)
from plotpy.styles.polygonmap import PolygonMapParam, PolygonParam_MS
from plotpy.styles.shape import (
    AnnotationParam,
    AnnotationParam_MS,
    AxesShapeParam,
    MarkerParam,
    RangeShapeParam,
    ShapeParam,
)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""PlotTypes test"""

# guitest: show

from guidata.qthelpers import qt_app_context
from numpy import linspace, sin

from plotpy.builder import make
from plotpy.tests import data as ptd
from plotpy.tests import vistools as ptv


def test_plot_types():
    """Test plot types"""
    _persist = []
    with qt_app_context(exec_loop=True):
        x = linspace(-10, 10, 200)
        y = sin(sin(sin(x)))
        for plot_type in ("curve", "image", "auto", "manual"):
            for reverse in (True, False):
                revtxt = f"{'image' if reverse else 'curve'} added first"
                items = [
                    make.curve(x, y, color="b"),
                    make.image(ptd.gen_image1()),
                ]
                _persist.append(
                    ptv.show_items(
                        reversed(items) if reverse else items,
                        plot_type=plot_type,
                        wintitle=f"Plot type: '{plot_type}' ({revtxt})",
                    )
                )


if __name__ == "__main__":
    test_plot_types()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
plotpy benchmarks
=================
"""

from .test_benchmarks import test_benchmarks


def run() -> None:
    """Run plotpy benchmarks"""
    test_benchmarks()


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Testing plot with synchronized axes"""

from __future__ import annotations

import numpy as np
from guidata.qthelpers import qt_app_context
from qtpy import QtGui as QG

from plotpy.builder import make
from plotpy.plot import BasePlot, PlotOptions
from plotpy.plot.plotwidget import SyncPlotWindow
from plotpy.tests.data import gen_2d_gaussian


def plot(plot_type, *itemlists):
    """Plot items in SyncPlotDialog"""
    win = SyncPlotWindow(
        title="Window for showing plots, optionally synchronized",
        options=PlotOptions(type=plot_type),
    )
    row, col = 0, 0
    for items in itemlists:
        plot = BasePlot()
        for item in items:
            plot.add_item(item)
        plot.set_axis_font("left", QG.QFont("Courier"))
        plot.set_items_readonly(False)
        win.add_plot(row, col, plot, sync=True)
        col += 1
        if col == 2:
            row += 1
            col = 0
    win.finalize_configuration()
    if plot_type == "image":
        win.manager.get_contrast_panel().show()
    win.resize(800, 600)
    win.show()


def test_syncplot_curves():
    """Test plot synchronization: curves"""
    x = np.linspace(-10, 10, 200)
    dy = x / 100.0
    y = np.sin(np.sin(np.sin(x)))
    x2 = np.linspace(-10, 10, 20)
    y2 = np.sin(np.sin(np.sin(x2)))
    with qt_app_context(exec_loop=True):
        plot(
            "curve",
            [
                make.curve(x, y, color="b"),
                make.label(
                    "Relative position <b>outside</b>", (x[0], y[0]), (-10, -10), "BR"
                ),
            ],
            [make.curve(x2, y2, color="g")],
            [
                make.curve(x, np.sin(2 * y), color="r"),
                make.label(
                    "Relative position <i>inside</i>", (x[0], y[0]), (10, 10), "TL"
                ),
            ],
            [
                make.merror(x, y / 2, dy),
                make.label("Absolute position", "R", (0, 0), "R"),
                make.legend("TR"),
            ],
        )


def test_syncplot_images():
    """Test plot synchronization: images"""
    img1 = gen_2d_gaussian(20, np.uint8, x0=-10, y0=-10, mu=7, sigma=10.0)
    img2 = gen_2d_gaussian(20, np.uint8, x0=-10, y0=-10, mu=5, sigma=8.0)
    img3 = gen_2d_gaussian(20, np.uint8, x0=-10, y0=-10, mu=3, sigma=6.0)
    with qt_app_context(exec_loop=True):

        def makeim(data):
            """Make image item"""
            return make.image(data, interpolation="nearest")

        plot("image", [makeim(img1)], [makeim(img2)], [makeim(img3)])


if __name__ == "__main__":
    test_syncplot_curves()
    test_syncplot_images()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
pyplot test

Interactive plotting interface with MATLAB-like syntax
"""

# guitest: show

import numpy as np
import pytest
from numpy.random import normal

import plotpy.pyplot as plt


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_pyplot():
    x = np.linspace(-5, 5, 1000)
    plt.figure(1)
    plt.subplot(2, 1, 1)
    plt.plot(x, np.sin(x), "r+")
    plt.plot(x, np.cos(x), "g-")
    plt.errorbar(x, -1 + x**2 / 20 + 0.2 * np.random.rand(len(x)), x / 20)
    plt.xlabel("Axe x")
    plt.ylabel("Axe y")
    plt.subplot(2, 1, 2)
    img = np.fromfunction(
        lambda x, y: np.sin((x / 200.0) * (y / 200.0) ** 2), (1000, 1000)
    )
    plt.xlabel("pixels")
    plt.ylabel("pixels")
    plt.zlabel("intensity")
    plt.gray()  # pylint: disable=no-member
    plt.imshow(img)
    #    savefig("D:\\test1.pdf", draft=True)

    plt.figure("table plot")
    data = np.array([x, np.sin(x), np.cos(x)]).T
    plt.plot(data)

    plt.figure("simple plot")
    plt.subplot(1, 2, 1)
    plt.plot(x, np.tanh(x + np.sin(12 * x)), "g-", label="Tanh")
    plt.legend()
    plt.subplot(1, 2, 2)
    plt.plot(x, np.sinh(x), "r:", label="SinH")
    #    savefig("D:\\test2.pdf")
    #    savefig("D:\\test2.png")
    plt.show()

    plt.figure("semilogx")
    plt.semilogx(x, np.sin(12 * x), "g-")
    plt.show()

    plt.figure("plotyy")
    plt.plotyy(x, np.sin(x), x, np.cos(x))
    plt.ylabel("sinus", "cosinus")
    plt.show()

    plt.figure("hist")

    data = normal(0, 1, (2000,))
    plt.hist(data)
    plt.show()

    plt.figure("pcolor 1")
    r = np.linspace(1.0, 16, 100)
    th = np.linspace(0.0, np.pi, 100)
    R, TH = np.meshgrid(r, th)
    X = R * np.cos(TH)
    Y = R * np.sin(TH)
    Z = 4 * TH + R
    plt.pcolor(X, Y, Z)

    plt.figure("pcolor 2")
    plt.pcolor(Z)
    plt.hot()  # pylint: disable=no-member
    plt.show()


if __name__ == "__main__":
    test_pyplot()



from .sliders import (
    QDoubleRangeSlider,
    QDoubleSlider,
    QLabeledDoubleRangeSlider,
    QLabeledDoubleSlider,
    QLabeledRangeSlider,
    QLabeledSlider,
    QRangeSlider,
)

__all__ = [
    "QDoubleRangeSlider",
    "QDoubleSlider",
    "QLabeledDoubleRangeSlider",
    "QLabeledDoubleSlider",
    "QLabeledRangeSlider",
    "QLabeledSlider",
    "QRangeSlider",
]



# -*- coding: utf-8 -*-
"""
about
=====

"""

from __future__ import annotations

import guidata
import qwt
from guidata.configtools import get_icon
from guidata.widgets import about as guidata_about
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMainWindow, QMessageBox

import plotpy
from plotpy.config import _


def get_python_libs_infos(addinfos: str = "") -> str:
    """Get Python and libraries information

    Args:
        addinfos: additional information to be displayed

    Returns:
        str: Python and libraries information
    """
    if addinfos:
        addinfos = ", " + addinfos
    addinfos = f"PlotPy {plotpy.__version__}{addinfos}"
    return guidata_about.get_python_libs_infos(addinfos)


def about(html: bool = True, copyright_only: bool = False) -> str:
    """Return text about this package

    Args:
        html: return html text. Defaults to True.
        copyright_only: if True, return only copyright

    Returns:
        str: text about this package
    """
    info = guidata_about.AboutInfo(
        name="PlotPy",
        version=plotpy.__version__,
        description=_("Set of tools for curve and image plotting"),
        author="Pierre Raybaut",
        year=2016,
        organization="PlotPyStack",
    )
    addinfos = f"guidata {guidata.__version__}, PythonQwt {qwt.__version__}"
    return info.about(html=html, copyright_only=copyright_only, addinfos=addinfos)


def show_about_dialog() -> None:
    """Show ``plotpy`` about dialog"""
    win = QMainWindow(None)
    win.setAttribute(Qt.WA_DeleteOnClose)
    win.hide()
    win.setWindowIcon(get_icon("plotpy.svg"))
    QMessageBox.about(win, _("About") + " PlotPy", about(html=True))
    win.close()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test showing a pcolor plot"""

# guitest: show

from time import time

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy._scaler import _scale_quads
from plotpy.builder import make
from plotpy.tests import vistools as ptv


def polar_demo(N=300):
    r = np.linspace(1.0, 16, N)
    th = np.linspace(0.0, np.pi, N)

    R, TH = np.meshgrid(r, th)
    X = R * np.cos(TH)
    Y = R * np.sin(TH)
    Z = 4 * TH + R + np.random.randint(-8, 8, size=(N, N))
    ix = np.random.randint(N, size=(N // 20,))
    iy = np.random.randint(N, size=(N // 20,))
    Z[ix, iy] = np.nan
    return X, Y, Z


def compute_quads(N=300):
    X, Y, Z = polar_demo(N)
    item = make.pcolor(X, Y, Z)
    return [item]


def compute_quads2(N=4):
    x = np.linspace(0, N - 1, N)
    X, Y = np.meshgrid(x, x)
    Z = X**2 + Y**2
    Z = Z.reshape((8, 2))
    item = make.pcolor(Z)  # checks if X, Y are properly generated in make.pcolor
    return [item]


def compute_quads3():
    pi = np.pi
    cos = np.cos
    sin = np.sin
    items = []
    for i, t in enumerate(np.linspace(0, 2 * pi, 16)):
        X = np.array([[0.0, cos(t)], [-sin(t), cos(t) - sin(t)]])
        Y = np.array([[0.0, sin(t)], [cos(t), sin(t) + cos(t)]])
        Z = np.array([[1.0, 2.0], [3.0, 4.0]])
        item = make.pcolor(X - 16 + 2 * i, Y - 3, Z)
        items.append(item)
    return items


def test_pcolor():
    """Testing pcolor"""
    with qt_app_context(exec_loop=True):
        items = compute_quads() + compute_quads3()
        _win = ptv.show_items(items, plot_type="image", wintitle=test_pcolor.__doc__)


def test_valgrind(K=20):
    X, Y, Z = polar_demo()
    lut = (1.0, 0.0, None, np.zeros((1024,), np.uint32))
    offscreen = np.zeros((1200, 1920), np.uint32)
    border = 1
    interpolate = (0, 0.5, 0.5)
    src_rect = (X.min(), Y.min(), X.max(), Y.max())
    dst_rect = (0, 0, 1920, 1200)
    t0 = time()
    for _count in range(K):
        _scale_quads(X, Y, Z, src_rect, offscreen, dst_rect, lut, interpolate, border)
    print((time() - t0) / K, " secs per loop")


if __name__ == "__main__":
    test_pcolor()
    test_valgrind(K=200)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Curve plotting test"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import vistools as ptv


def test_plot():
    """Curve plotting test"""
    x = np.linspace(-10, 10, 200)
    dy = x / 100.0
    y = np.sin(np.sin(np.sin(x)))
    x2 = np.linspace(-10, 10, 20)
    y2 = np.sin(np.sin(np.sin(x2)))
    with qt_app_context(exec_loop=True):
        items = [
            make.curve(x, y, color="b"),
            make.curve(x2, y2, color="g", curvestyle="Sticks", title="toto"),
            make.curve(x, np.sin(2 * y), color="r"),
            make.merror(x, y / 2, dy),
            make.label(
                "Relative position <b>outside</b>", (x[0], y[0]), (-10, -10), "BR"
            ),
            make.label("Relative position <i>inside</i>", (x[0], y[0]), (10, 10), "TL"),
            make.label("Absolute position", "R", (0, 0), "R"),
            make.legend("TR"),
            make.marker(
                position=(5.0, 0.8),
                label_cb=lambda x, y: "A = %.2f" % x,
                markerstyle="|",
                movable=False,
            ),
        ]
        _win = ptv.show_items(
            items, wintitle=test_plot.__doc__, title="Curves", plot_type="curve"
        )


if __name__ == "__main__":
    test_plot()
